#ifndef __DRIVER_H__
#define __DRIVER_H__

#include "TuxLCD.h"
using namespace std;

/**
 * Generic interface for TuxLCD drivers.
 *
 * \author David Bull
 * \version 0.1
 * \date Monday 7th August 2006
 */
class Driver
{
	private:
	
	public:	
		/**
		 * Destrustor.
		 */
		virtual ~Driver() { }
		
		
		/**
		 * Returns the number of columns the screen has.
		 * \return the number of columns the screen has.
		 */
		virtual int getScreenColumns() = 0;
		
		
		/**
		 * Returns the number of rows the screen has.
		 * \return the number of rows the screen has.
		 */
		virtual int getScreenRows() = 0;
		
		
		/**
		 * Returns the number of GPOs the screen has.
		 * \return the number of GPOs the screen has.
		 */
		virtual int getScreenGPOs() = 0;
		
		
		/**
		 *
		 * \param[in] buffer
		 * \param[in] length
		 * \return
		 */
		virtual int sendBuffer(char *buffer, int length) = 0;
		
		
		/**
		 * Clears the display and resets cursor position to top left.
		 * \return
		 */
		virtual int clearScreen() = 0;
		
		
		/**
		 * This command sets the display's contrast.
		 * \param[in] contrast
		 * \return
		 */
		virtual int setContrast(unsigned char contrast) = 0;
		
		
		/**
		 * Returns the contrast of the display.
		 * \return the contrast of the display.
		 */
		virtual unsigned char getContrast() = 0;
		
		
		/**
		 * This commands sets the display's backlight brightness.
		 * \param[in] brightness
		 * \return
		 */
		virtual int setBrightness(unsigned char brightness) = 0;
		
		
		/**
		 * Returns the brightness of the display.
		 * \return the brightness of the display.
		 */
		virtual unsigned char getBrightness() = 0;
		
		
		/**
		 * This command turns on the backlight for a time of [minutes] minutes.
		 * If [minutes] is zero (0), the backlight will remain on indefinitely.
		 * \param[in] minutes
		 * \return
		 */
		virtual int backlightOn(char minutes) = 0;
		
		
		/**
		 * This command turns the backlight off.
		 * \return
		 */
		virtual int backlightOff() = 0;
		
		
		/**
		 * Toggles the backlight.
		 * \return
		 */
		virtual int toggleBacklight() = 0;
		
		
		/**
		 * Returns the state of backlight.
		 * \return the state of backlight.
		 */
		virtual bool getBacklightState() = 0;
		
		
		/**
		 * This command turns ON or OFF any of the general purpose outputs.
		 * \param[in] gpo
		 * \param[in] enabled
		 * \return
		 */
		virtual int enableGpo(char gpo, bool enabled) = 0;
		
		
		/**
		 * Enables or diables automatic line wrapping.
		 * \param[in] enabled
		 * \return
		 */
		virtual int enableAutoLineWrap(bool enabled) = 0;
		
		
		/**
		 * Enables or disables automatic scrolling.
		 * \param[in] enabled
		 * \return
		 */
		virtual int enableAutoScroll(bool enabled) = 0;
		
		
		/**
		 * This command sets the cursor position (text insertion point) to the [column] and [row] specified.
		 * \param[in] column
		 * \param[in] row
		 * \return
		 */
		virtual int setCursorPosition(char column, char row) = 0;
		
		
		/**
		 * This command moves the cursor position (text insertion point) to the top left of the display area.
		 * \return
		 */
		virtual int sendCursorHome() = 0;
		
		
		/**
		 * Turns on or off the underline cursor.
		 * \param[in] enabled
		 * \return
		 */
		virtual int enableUnderlineCursor(bool enabled) = 0;
		
		
		/**
		 * Turns on or off the blinking block cursor.
		 * \param[in] enabled
		 * \return
		 */
		virtual int enableBlinkingCursor(bool enabled) = 0;
		
		
		/**
		 * Moves the cursor one position to the left but does not erase any character that may be in that position.
		 * \return
		 */
		virtual int cursorLeft() = 0;
		
		
		/**
		 * Moves the cursor one position to the right but does not erase any character that may be in that position.
		 * \return
		 */
		virtual int cursorRight() = 0;
		
		
		/**
		 * Defines a custom characters.
		 * \param[in] charNum 0-7
		 * \param[in] charCata 8 byte character map
		 * \return
		 */
		virtual int defineCustomChar(char charNum, const char charCata[8]) = 0;
		
		
		/**
		 * Defines 8 custom characters.
		 * \param[in] customCharSet
		 * \return
		 */
		virtual int defineCustomCharSet(const char customCharSet[8][8]) = 0;
};

// the types of the class factories
typedef Driver* create_driver_t(void*);
typedef void destroy_driver_t(Driver*);

#endif
