#include <cstdlib>		// Standard library definitions
#include <unistd.h>		// UNIX standard function definitions
#include <cstdio>		// Standard input/output definitions
#include <iostream>		// Classes for input and output operations
#include <sys/stat.h>		//
#include <fstream>		//
#include <string>		// Standard library to manipulate C strings#include <iostream>
#include <sstream>		//
#include <vector>		//
#include <csignal>		//
#include <time.h>		//
#include <dlfcn.h>		//
#include <glib.h>		//
#include <popt.h>		// Command line option processing

#include "TuxLCD.h"
#include "TuxLCDException.h"
#include "Plugin.h"

using namespace std;


bool tuxlcdRunning;


/**
 * Constructor.
 * \param[in] configFile full path to TuxLCD configuration file.
 * \param[in] foreground if true then TuxLCD stays in the foreground, otherwise it runs as a deamon.
 */
TuxLCD::TuxLCD(const char *configFile, bool foreground) :
	lcdDriver(NULL), driverName(NULL), screenConfig(NULL), eventConfig(NULL), driverDir(NULL), pluginDir(NULL), customCharDir(NULL)
{
	log(LOG_LEVEL_DEBUG, "TuxLCD: Constructing...\n");
	this->foreground = foreground;
	loadConfig(configFile);
}


/**
 * Initialises TuxLCD.
 * \return 0 on success and 1 on failure.
 */
int TuxLCD::init()
{
	char *filePath = NULL;
	try
	{
		log(LOG_LEVEL_DEBUG, "TuxLCD: Creating LK202 driver...\n");
		filePath = g_strconcat(driverDir, "Driver_", driverName, ".so", NULL);
		loadDriver(filePath);
		free(filePath);
	}
	catch (TuxLCDException *e)
	{
		if (filePath!=NULL)
			free(filePath);
		cerr << e->what();
		return 1;
	}
	return 0;
}


/**
 * Destrustor.
 * Displays a goodbye message on the LCD, unloads the screens and plugins, and destroys the TuxLCD driver.
 */
TuxLCD::~TuxLCD()
{
	log(LOG_LEVEL_DEBUG, "TuxLCD: Destructing...\n");
	if (lcdDriver!=NULL)
	{
		int bufLen = lcdDriver->getScreenColumns();
		char *buffer = (char *)calloc(bufLen, sizeof(char));
		char *tmp;
		
		lcdDriver->clearScreen();
		lcdDriver->setContrast(displayContrast);
		centerText("Thanks for using", bufLen, buffer);
		lcdDriver->sendBuffer(buffer, bufLen);
		lcdDriver->setCursorPosition(1, 2);
		tmp = g_strconcat("TuxLCD - ", getenv("LOGNAME"), NULL);
		centerText(tmp, bufLen, buffer);
		lcdDriver->sendBuffer(buffer, bufLen);
		sleep(1);
		
		unloadEvents();
		unloadScreens();
		unloadPlugins();
		destroy_driver(lcdDriver);
		free(buffer);
		free(tmp);
	}
}


/**
 * Main TuxLCD method.
 * Initialises the TuxLCD Driver, loads the plugins and screen definitions,
 * and then loops forever displaying the screens.
 */
void TuxLCD::main()
{
	log(LOG_LEVEL_DEBUG, "TuxLCD: Starting...\n");
	tuxlcdRunning = true;
	int retval;
	char *filePath = NULL;
	char myCustomCharSet[8][8];
	int bufLen;
	char *buffer1;
	char *buffer2;
	
	mainThread = pthread_self();
	signal(SIGINT, tuxlcdSignalHandler);
	signal(SIGTERM, tuxlcdSignalHandler);
	signal(SIGUSR1, tuxlcdSignalHandler);
	
	displayContrast = lcdDriver->getContrast();
	bufLen = lcdDriver->getScreenColumns();
	buffer1 = (char *)calloc(bufLen, sizeof(char));
	buffer2 = (char *)calloc(bufLen, sizeof(char));
	loadPlugins(pluginDir);
	loadScreens(screenConfig);
	loadEvents(eventConfig);
	
	
	/*
	 * Display the TuxLCD title screen
	 */
	log(LOG_LEVEL_DEBUG, "TuxLCD: Displaying title screen...\n");
	lcdDriver->clearScreen();
	filePath = g_strconcat(customCharDir, "Big Lightning.lcdc-cc", NULL);
	loadCustomCharSet(filePath, myCustomCharSet);
	free(filePath);
	lcdDriver->defineCustomCharSet(myCustomCharSet);
	buffer1[0] = 2;
	buffer1[1] = 3;
	buffer1[2] = ' ';
	memcpy(buffer2, "TuxLCD ", 7);
	memcpy(buffer2+7, VERSION, strlen(VERSION));
	centerText(buffer2, bufLen-3, buffer1+3);
	buffer2[0] = 6;
	buffer2[1] = 7;
	buffer2[2] = ' ';
	centerText("(c) uk_dave 2006", bufLen-3, buffer2+3);
	lcdDriver->sendBuffer(buffer1, bufLen);
	lcdDriver->setCursorPosition(1, 2);
	lcdDriver->sendBuffer(buffer2, bufLen);
	sleep(3);
	
	
	Screen screen;
	screenIndex = 0;
	int bufLen1, bufLen2;
	while (tuxlcdRunning)
	{
		screen = screens[screenIndex];
		if (strlen(screen.customChars)>0)
		{
			filePath = g_strconcat(customCharDir, screen.customChars, ".lcdc-cc", NULL);
			loadCustomCharSet(filePath, myCustomCharSet);
			free(filePath);
		}
		bufLen1 = parseScreenLine(screen.line1, buffer1, 20);
		bufLen2 = parseScreenLine(screen.line2, buffer2, 20);
displayContrast = lcdDriver->getContrast();
transitionFadeOut();
		lcdDriver->clearScreen();
		lcdDriver->defineCustomCharSet(myCustomCharSet);
		lcdDriver->sendBuffer(buffer1, bufLen1);
		lcdDriver->setCursorPosition(1, 2);
		lcdDriver->sendBuffer(buffer2, bufLen2);
transitionFadeIn(displayContrast);
		retval = sleep(screen.duration);
		if (retval==0)
		{
			screenIndex++;
			if (screenIndex>=screens.size())
				screenIndex = 0;
		}
	}
	
	free(buffer1);
	free(buffer2);
}


/**
 * Prints the properties for the specified plugin to the console.
 * If the specified plugin was not found a list of all available plugins is printed.
 * \param[in] pluginName the name of the plugin.
 */
void TuxLCD::printPluginProps(const char *pluginName)
{
	int i=0;
	bool foundPlugin = false;
	loadPlugins(pluginDir);
	for (i=0; i<plugins.size(); i++)
	{
		if (strcmp(plugins[i].plugin_instance->getPluginName(), pluginName) == 0)
		{
			 plugins[i].plugin_instance->dumpProperties();
			 foundPlugin = true;
			 break;
		}
	}
	if (foundPlugin == false)
	{
		cout << "Plugin '" << pluginName << "' not found. Available plugins are:" << endl;
		for (i=0; i<plugins.size(); i++)
			cout << "  " << plugins[i].plugin_instance->getPluginName() << endl;
	}
	unloadPlugins();
}


/**
 * Signal handler for SIGINT, SIGTERM, and SIGUSR1.
 * We trap SIGINT and SIGTERM so that we can perform a nice shutdown of TuxLCD.
 * SIGUSR1 is used to interrupt the sleep() between screens so that, for example, we can skip between screens.
 * \param[in] signum the signal number.
 */
void tuxlcdSignalHandler(int signum)
{
//	log(LOG_LEVEL_DEBUG, "TuxLCD: tuxlcdSignalHandler() caught a signal\n");
	if (signum==SIGINT || signum==SIGTERM)
		tuxlcdRunning = false;
}


/**
 * Called from a TuxLCD Driver when a key on the screens keypad is pressed.
 * \param[in] key the key that was pressed.
 */
void TuxLCD::lcdKeyPressed(char key)
{
	string msg = "TuxLCD: key pressed = ";
	msg.append(&key);
	msg.append("\n");
	log(LOG_LEVEL_DEBUG, msg);
	if (key=='H')		// Enter
	{
		lcdDriver->toggleBacklight();
	}
	else if (key=='O')	// Left
	{
		if (screenIndex == 0)
			screenIndex = screens.size()-1;
		else
			screenIndex--;
		pthread_kill(mainThread, SIGUSR1);
	}
	else if (key=='E')	// Right
	{
		screenIndex++;
		if (screenIndex>=screens.size())
			screenIndex = 0;
		pthread_kill(mainThread, SIGUSR1);
	}
}


/**
 *
 * \param[in] contrast
 */
void TuxLCD::transitionFadeIn(unsigned char contrast)
{
	unsigned char i;
	timespec req, rem;
	req.tv_sec = 0;
	req.tv_nsec = 500000/contrast;
	for (i=0; i<=contrast; i++)
	{
		lcdDriver->setContrast(i);
		nanosleep(&req, &rem);
	}
}


/**
 *
 */
void TuxLCD::transitionFadeOut()
{
	unsigned char i, contrast;
	timespec req, rem;
	contrast = lcdDriver->getContrast();
	req.tv_sec = 0;
	req.tv_nsec = 500000/contrast;
	for (i=contrast; i>0; i--)
	{
		lcdDriver->setContrast(i);
		nanosleep(&req, &rem);
	}
}


/**
 * Parses as line of text from a screen definition replacing all tokens with the values from the corresponding plugins.
 * Custom character tokens are defined as [$CCx] where 'x' is 0-7.
 * Plugin tokens are defined as [$name.property] where 'name' is the name of the plugin and 'property' is a property from the specified plugin.
 * NOTE: If the buffer is not long enough to contain the processed output then the output is truncated.
 * \param[in] line the line to parse.
 * \param[in,out] buffer the buffer in which to place the processed output.
 * \param[in] bufferLen the length of the output buffer.
 * \return the number of characters placed in the output buffer.
 */
int TuxLCD::parseScreenLine(char *line, char *buffer, size_t bufferLen)
{
	size_t lineLength = strlen(line);
	char *linePtr = line;
	char *endDelim = line;
	char *bufferPtr = buffer;
	
	char *dotPtr;
	char *plugName = NULL;
	char *propName = NULL;
	char *propVal;
	size_t i;
	
	while ((linePtr-line < lineLength) && (bufferPtr-buffer < bufferLen))
	{
		if (strncmp(linePtr, "[$", 2)==0)
		{
			endDelim = strchr(linePtr, ']');
			if (endDelim == NULL)
			{
				string msg = "TuxLCD: ERROR: parseScreenLine(): Screen line is invalid '";
				msg.append(line);
				msg.append("\n");
				log(LOG_LEVEL_ERROR, msg);
				break;
			}
			if (endDelim-linePtr==5 && *(linePtr+2)=='C' && *(linePtr+3)=='C' && *(linePtr+4)>=48 && *(linePtr+4)<=55)
			{
				*bufferPtr = *(linePtr+4) - 48;
				bufferPtr++;
			}
			else
			{
				dotPtr = strchr(linePtr, '.');
				if (dotPtr != NULL)
				{
					plugName = (char *)realloc(plugName, dotPtr-linePtr-1);
					memcpy(plugName, linePtr+2, dotPtr-linePtr-2);
					plugName[dotPtr-linePtr-2] = 0;
					
					propName = (char *)realloc(propName, endDelim-dotPtr);
					memcpy(propName, dotPtr+1, endDelim-dotPtr-1);
					propName[endDelim-dotPtr-1] = 0;
					
					for (i=0; i<plugins.size(); i++)
					{
						if (strcmp(plugins[i].plugin_instance->getPluginName(), plugName) == 0)
						{
							propVal = plugins[i].plugin_instance->getProperty(propName);
							if (propVal != NULL)
							{
								if (strlen(propVal)+(bufferPtr-buffer) < bufferLen)
								{
									memcpy(bufferPtr, propVal, strlen(propVal));
									bufferPtr += strlen(propVal);
								}
								else
								{
									memcpy(bufferPtr, propVal, bufferLen-(bufferPtr-buffer));
									bufferPtr += bufferLen-(bufferPtr-buffer);
								}
								free(propVal);
							}
							break;
						}
					}
				}
				else
				{
					*bufferPtr = *linePtr;
					bufferPtr++;
					linePtr++;
				}
				
			}
			linePtr = endDelim+1;
		}
		else
		{
			*bufferPtr = *linePtr;
			bufferPtr++;
			linePtr++;
		}
	}
	free(plugName);
	free(propName);
	return bufferPtr - buffer;
}


/**
 * Loads the main TuxLCD configuration file.
 * \param[in] configFile the full path to the configuration file.
 * \throw TuxLCDException if the config file could not be opened or an option was missing.
 */
void TuxLCD::loadConfig(const char *configFile)
{
	string msg = "TuxLCD: Loading configuration from '";
	msg.append(configFile);
	msg.append("'...\n");
	log(LOG_LEVEL_INFO, msg);
	ifstream fin;
	string str;
	bool mainConfigSectionFound = false;
	
	fin.open(configFile, ios::in);
	if (fin.is_open() == false)
	{
		msg = "TuxLCD: ERROR: Unable to open configuration file '";
		msg.append(configFile);
		msg.append("'\n");
		log(LOG_LEVEL_FATAL, msg);
		throw new TuxLCDException(msg.c_str());
	}
	while (fin.eof() == false)
	{
		getline(fin, str);
		
		if (str.length()>3 && str.at(0)=='[')
		{
			if (str.length()==13 && strncmp(str.data(), "[Main_TuxLCD]", 13)==0)
				mainConfigSectionFound = true;
		}
		
		if (mainConfigSectionFound == true)
		{
			if (str.length()>7 && strncmp(str.data(), "Driver=", 7)==0)
				driverName = strdup(str.substr(7).c_str());
			if (str.length()>8 && strncmp(str.data(), "Screens=", 8)==0)
				screenConfig = strdup(str.substr(8).c_str());
			if (str.length()>7 && strncmp(str.data(), "Events=", 7)==0)
				eventConfig = strdup(str.substr(7).c_str());
			if (str.length()>10 && strncmp(str.data(), "DriverDir=", 10)==0)
				driverDir = strdup(str.substr(10).c_str());
			if (str.length()>10 && strncmp(str.data(), "PluginDir=", 10)==0)
				pluginDir = strdup(str.substr(10).c_str());
			if (str.length()>14 && strncmp(str.data(), "CustomCharDir=", 14)==0)
				customCharDir = strdup(str.substr(14).c_str());
		}
	}
	fin.close();
	
	if (mainConfigSectionFound == false)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: Can't find [Main_TuxLCD] section in config file.\n");
		throw new TuxLCDException("Error can't find [Main_TuxLCD] section in config file.");
	}
	if (driverName == NULL)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: 'driverName' option missing in config file.\n");
		throw new TuxLCDException("Error 'driverName' option missing in config file.");
	}
	if (screenConfig == NULL)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: 'screenConfig' option missing in config file.\n");
		throw new TuxLCDException("Error 'screenConfig' option missing in config file.");
	}
	if (eventConfig == NULL)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: 'eventConfig' option missing in config file.\n");
		throw new TuxLCDException("Error 'eventConfig' option missing in config file.");
	}
	if (driverDir == NULL)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: 'driverDir' option missing in config file.\n");
		throw new TuxLCDException("Error 'driverDir' option missing in config file.");
	}
	if (pluginDir == NULL)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: 'pluginDir' option missing in config file.\n");
		throw new TuxLCDException("Error 'pluginDir' option missing in config file.");
	}
	if (customCharDir == NULL)
	{
		log(LOG_LEVEL_FATAL, "TuxLCD: ERROR: 'customCharDir' option missing in config file.\n");
		throw new TuxLCDException("Error 'customCharDir' option missing in config file.");
	}
}


/**
 * Loads the specified TuxLCD display driver.
 * \param[in] fileName full path to driver file.
 * \throw TuxLCDException if there was a problem loading the driver.
 */
void TuxLCD::loadDriver(const char *fileName)
{
	string msg = "TuxLCD: Loading driver '";
	msg.append(fileName);
	msg.append("'...");
	log(LOG_LEVEL_INFO, msg);
	
	const char *dlsym_error;
	void *lib;
	
	lib = dlopen(fileName, RTLD_NOW);
	if (!lib)
	{
		log(LOG_LEVEL_INFO, "FAILED\n");
		msg = "TuxLCD:   ERROR: Cannot load driver: ";
		msg.append(dlerror());
		msg.append("\n");
		log(LOG_LEVEL_FATAL, msg);
		throw new TuxLCDException(msg.c_str());
	}
	else
	{
		dlerror();
		create_driver = (create_driver_t*)dlsym(lib, "create");
		dlsym_error = dlerror();
		if (dlsym_error)
		{
			log(LOG_LEVEL_INFO, "FAILED\n");
			msg = "TuxLCD:   ERROR: Cannot load symbol 'create': ";
			msg.append(dlsym_error);
			msg.append("\n");
			log(LOG_LEVEL_FATAL, msg);
			throw new TuxLCDException(msg.c_str());
		}
		else
		{
			destroy_driver = (destroy_driver_t*)dlsym(lib, "destroy");
			dlsym_error = dlerror();
			if (dlsym_error)
			{
				log(LOG_LEVEL_INFO, "FAILED\n");
				msg = "TuxLCD:   ERROR: Cannot load symbol 'destroy': ";
				msg.append(dlsym_error);
				msg.append("\n");
				log(LOG_LEVEL_FATAL, msg);
				throw new TuxLCDException(msg.c_str());
			}
			else
			{
				log(LOG_LEVEL_INFO, "OK\n");
				lcdDriver = create_driver(this);
			}
		}
	}
}


/**
 * Loads the plugins.
 * The plugins directory is scanned and all shared objects are loaded.
 * An instance of each of the plugins is created and places in the plugins
 * vector along with pointers to its create and destroy functions.
 * \param[in] pluginDir the directory containing the plugins.
 */
void TuxLCD::loadPlugins(const char *pluginDir)
{
	log(LOG_LEVEL_INFO, "TuxLCD: Loading plugins...\n");
	GDir *dir;
	gchar *name, *pluginpath, *dot;
	string msg;
	
	const char *dlsym_error;
	void *lib;
	create_plugin_t *create_plugin;
	destroy_plugin_t *destroy_plugin;
	
	dir = g_dir_open(pluginDir, 0, NULL);
	if (dir == NULL)
	{
		msg = "TuxLCD: ERROR: Unable to open plugin directory '";
		msg.append(pluginDir);
		msg.append("\n");
		log(LOG_LEVEL_ERROR, msg);
		return;
	}
	
	while (1)
	{
		name = (gchar *)g_dir_read_name(dir);
		if (name == NULL)
			break;
		dot = strrchr(name, '.');
		if (!(dot && strcmp(dot+1, "so")==0))
			continue;
		
		msg = "TuxLCD: Loading plugin '";
		msg.append(name);
		msg.append("'... ");
		log(LOG_LEVEL_INFO, msg);
		
		pluginpath = g_strconcat(pluginDir, name, NULL);
		lib = dlopen(pluginpath, RTLD_NOW);
		if (!lib)
		{
			log(LOG_LEVEL_INFO, "FAILED\n");
			msg = "TuxLCD:   ERROR: Cannot load plugin: ";
			msg.append(dlerror());
			msg.append("\n");
			log(LOG_LEVEL_ERROR, msg);
			continue;
		}
		dlerror();
		create_plugin = (create_plugin_t*)dlsym(lib, "create");
		dlsym_error = dlerror();
		if (dlsym_error)
		{
			log(LOG_LEVEL_INFO, "FAILED\n");
			msg = "TuxLCD:   ERROR: Cannot load symbol 'create': ";
			msg.append(dlsym_error);
			msg.append("\n");
			log(LOG_LEVEL_FATAL, msg);
			continue;
		}
		destroy_plugin = (destroy_plugin_t*)dlsym(lib, "destroy");
		dlsym_error = dlerror();
		if (dlsym_error)
		{
			log(LOG_LEVEL_INFO, "FAILED\n");
			msg = "TuxLCD:   ERROR: Cannot load symbol 'destroy': ";
			msg.append(dlsym_error);
			msg.append("\n");
			log(LOG_LEVEL_FATAL, msg);
			continue;
		}
		log(LOG_LEVEL_INFO, "OK\n");
		
		PluginData pluginData;
		pluginData.create_plugin = create_plugin;
		pluginData.plugin_instance = create_plugin(this);
		pluginData.destroy_plugin = destroy_plugin;
		plugins.push_back(pluginData);
		free(pluginpath);
	}
	
	g_dir_close(dir);
	ostringstream msgStream(ostringstream::out);
	msgStream << "TuxLCD: Loaded " << plugins.size() << " plugins" << endl;
	log(LOG_LEVEL_INFO, msgStream.str());
}


/**
 * Unloads the plugins.
 * Destroys the plugins and removes them from the plugins vector.
 */
void TuxLCD::unloadPlugins()
{
	log(LOG_LEVEL_INFO, "TuxLCD: Unloading plugins...\n");
	PluginData pluginData;
	
	while (!plugins.empty())
	{
		pluginData = plugins.back();
		pluginData.destroy_plugin(pluginData.plugin_instance);
		plugins.pop_back();
	}
}


/**
 * Loads the specified screen definition file.
 * Screens are processed into the Screen structure and placed into the screens vector.
 * \param[in] fileName full path to the screen definition file.
 */
void TuxLCD::loadScreens(char *fileName)
{
	string msg = "TuxLCD: Loading screens from '";
	msg.append(fileName);
	msg.append("'...\n");
	log(LOG_LEVEL_INFO, msg);
	ifstream fin;
	string str;
	
	fin.open(fileName, ios::in);
	if (fin.is_open() == false)
	{
		msg = "TuxLCD: ERROR: Unable to open screens file '";
		msg.append(fileName);
		msg.append("\n");
		log(LOG_LEVEL_ERROR, msg);
		return;
	}
	while (fin.eof() == false)
	{
		if (str.length()==0 || str.at(0)!='[')
			getline(fin, str);
		if (str.length()>10 && strncmp(str.data(), "[Display_", 9)==0 && str.at(str.length()-1)==']')
		{
			Screen screen;
			while (fin.eof() == false)
			{
				getline(fin, str);
				if (str.length()>5 && strncmp(str.data(), "Name=", 5)==0)
					screen.name = strdup((char *)str.substr(5).c_str());
				else if (str.length()>9 && strncmp(str.data(), "Duration=", 9)==0)
					screen.duration = atoi(str.substr(9).c_str());
				else if (str.length()>8 && strncmp(str.data(), "Enabled=", 8)==0)
					screen.enabled = (bool)atoi(str.substr(8).c_str());
				else if (str.length()>=12 && strncmp(str.data(), "CustomChars=", 12)==0)
				{
					if (str.length()==12)
						screen.customChars = strdup("");
					else
						screen.customChars = strdup((char *)str.substr(12).c_str());
				}
				else if (str.length()>13 && strncmp(str.data(), "Lines_0_Text=", 13)==0)
					screen.line1 = strdup((char *)str.substr(13).c_str());
				else if (str.length()>13 && strncmp(str.data(), "Lines_1_Text=", 13)==0)
					screen.line2 = strdup((char *)str.substr(13).c_str());
				else if (str.length()>1 && str.at(0)=='[')
					break;
			}
			screens.push_back(screen);
		}
	}
	fin.close();
	ostringstream msgStream(ostringstream::out);
	msgStream << "TuxLCD: Loaded " << screens.size() << " screens" << endl;
	log(LOG_LEVEL_INFO, msgStream.str());
}


/**
 * Unloads the screen definitions contained in the screens vector.
 * The memory used by the screen structures are freed and removed from the screens vector.
 */
void TuxLCD::unloadScreens()
{
	log(LOG_LEVEL_INFO, "TuxLCD: Unloading screens...\n");
	Screen screen;
	
	while (!screens.empty())
	{
		screen = screens.back();
		free(screen.name);
		free(screen.customChars);
		free(screen.line1);
		free(screen.line2);
		screens.pop_back();
	}
}


/**
 * TODO
 * \param[in] fileName
 */
void TuxLCD::loadEvents(char *fileName)
{
	string msg = "TuxLCD: Loading events from '";
	msg.append(fileName);
	msg.append("'... - ***NOT YET IMPLEMENTED***\n");
	log(LOG_LEVEL_INFO, msg);
	log(LOG_LEVEL_INFO, "TuxLCD: Loaded 0 events\n");
}


/**
 * TODO
 */
void TuxLCD::unloadEvents()
{
	log(LOG_LEVEL_INFO, "TuxLCD: Unloading events... - ***NOT YET IMPLEMENTED***\n");
}


/**
 * Loads an LCDC Custom Character file from disk into multidimensional char array.
 * The array must be char[8][8] and be allocated before being passed to this method.
 * \param[in] filename full path to the custom character file.
 * \param[in,out] customCharSet the char[8][8] array to load the custom characters into.
 * \return 0 if the custom character file was loaded correctly, 1 if not.
 */
int TuxLCD::loadCustomCharSet(const char *filename, char customCharSet[8][8])
{
	ifstream fin;
	string str;
	bool inCustomCharSection = false;
	char n, charNum, asciiNum;
	size_t indx;
	
	fin.open(filename, ios::in);
	if (fin.is_open() == false)
	{
		string msg = "TuxLCD: ERROR: Unable to open customchar file '";
		msg.append(filename);
		msg.append("\n");
		log(LOG_LEVEL_ERROR, msg);
		return 1;
	}
	
	for (charNum=0; charNum<=7; charNum++)
		for (n=0; n<=7; n++)
			customCharSet[charNum][n] = 0;
	
	while (fin.eof() == false)
	{
		getline(fin, str);
		if (str.size()<4)
			continue;
		if (str.at(0) == '[')
		{
			if (strncmp(str.data(), "[CustomCharacters]", 18) == 0)
				inCustomCharSection = true;
			else
				inCustomCharSection = false;
		}
		else if (inCustomCharSection == true)
		{
			if (str.at(0)>=48 && str.at(0)<=55 && str.at(1)=='=' && str.at(2)=='[')
			{
				charNum = str.at(0)-48;
				if (charNum>=0 && charNum<=7)
				{
					indx = 2;
					for (n=0; n<=7; n++)
					{
						if (indx+5 > str.size())
							break;
						if (str.at(indx)=='[' && str.at(indx+1)=='#' && str.at(indx+2)=='#' && str.at(indx+3)>=48 && str.at(indx+3)<=57)
						{
							asciiNum = str.at(indx+3)-48;
							if (str.at(indx+4)>=48 && str.at(indx+4)<=57)
							{
								asciiNum = (asciiNum*10) + (str.at(indx+4)-48);
								indx += 6;
							}
							else
								indx += 5;
							customCharSet[charNum][n] = asciiNum;
						}
					}
				}
			}
		}
	}
	fin.close();
	return 0;
}


/**
 * Centers the given text within the given buffer by prepending and appending spaces.
 * For example if we want to center "hello" into a buffer of size 9, the result will be "  hello  ".
 * If the length of the text is larger then the buffer then the text is simply copied to the buffer and the extra characters are cuttoff.
 * \param[in] text the text to center.
 * \param[in] centerLength the space withn which to center the text.
 * \param[in,out] centeredText the buffer that will contain the centered text.
 */
void TuxLCD::centerText(const char *text, size_t centerLength, char *centeredText)
{	
	unsigned int space, i;
	
	if (strlen(text) < centerLength)
	{
		space = centerLength - strlen(text);
		for (i=0; i<space/2; i++)
			*(centeredText+i) = ' ';
		memcpy(centeredText+(space/2), text, strlen(text));
		for (i=strlen(text)+(space/2); i<centerLength; i++)
			*(centeredText+i) = ' ';
	}
	else
	{
		memcpy(centeredText, text, centerLength);
	}
}


/**
 * Logs a message to either the console or a seperate log file depending on whether we are running as a daemon or not.
 * \param[in] level
 * \param[in] msg
 */
void TuxLCD::log(int level, string msg)
{
	if (foreground)
	{
		if (level <= LOG_LEVEL_ERROR)
			cerr << msg;
		else
			cout << msg;
	}
	else
	{
		// TODO
	}
}


bool checkIfAlreadyRunning()
{
	ifstream fin;
	fin.open(PID_FILE, ios::in);
	if (fin.is_open() == true)
	{
		string str;
		getline(fin, str);
		int pid = atoi(str.c_str());
		if (pid>0)
		{
			int retval = kill(pid, 0);
			if (retval > -1)
			{
				fin.close();
				return true;
			}
		}
	}
	fin.close();
	
	ofstream fout;
	fout.open(PID_FILE, ios::out);
	if (fout.is_open() == false)
	{
		cout << "Error opening PID file (" << PID_FILE << ") for writting" << endl;
		exit(EXIT_FAILURE);
	}
	fout << getpid();
	fout.close();
	
	return false;
}


/**
 *
 * \param argc
 * \param argv
 * \return
 */
int main(int argc, char *argv[])
{
	char	c;            /* used for argument parsing */
	char	*configFile = NULL;
	char	*pluginName = NULL;
	int	foreground = 0;
	int	version = 0;
	poptContext optCon;   /* context for parsing command-line options */
	
	struct poptOption optionsTable[] = {
		{ "config",       'c', POPT_ARG_STRING, &configFile, 0, "specify a config file (default: /etc/tuxlcd/tuxlcd.conf)", "CONFIG_FILE" },
		{ "foreground",   'f', POPT_ARG_NONE,   &foreground, 0, "stay in foreground (don't run as a daemon)" },
		{ "plugin_props", 'p', POPT_ARG_STRING, &pluginName, 0, "print plugin properties (specify '?' for a list of all plugins)", "PLUGIN_NAME" },
		{ "version",      'v', POPT_ARG_NONE,   &version,    0, "display version information" },
		POPT_AUTOHELP
		{ NULL, 0, 0, NULL, 0 }
	};
	optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
	while ((c = poptGetNextOpt(optCon)) >= 0);
	if (c < -1) 
	{
		/* an error occurred during option processing */
		fprintf(stderr, "%s: %s\n", poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
		return 1;
	}
	poptFreeContext(optCon);
	
	
	if (version)
	{
		cout << "TuxLCD version " << VERSION << " by uk_dave" << endl;
		cout << "Website: http://tuxlcd.sourceforge.net" << endl;
		return EXIT_SUCCESS;
	}
	
	if (configFile==NULL)
		configFile = DEFAULT_CONFIG;
	
	if (pluginName!=NULL)
	{
		TuxLCD *tuxlcd = new TuxLCD(configFile, foreground);
		tuxlcd->printPluginProps(pluginName);
		delete tuxlcd;
		return EXIT_SUCCESS;
	}
	
	if (!foreground)
	{
		/* Our process ID and Session ID */
		pid_t pid, sid;
		
		/* Fork off the parent process */
		pid = fork();
		if (pid < 0)
		{
			cout << "ERROR: fork failed" << endl;
			exit(EXIT_FAILURE);
		}
		/* If we got a good PID, then we can exit the parent process. */
		else if (pid > 0)
		{
			exit(EXIT_SUCCESS);
		}
		else
		{
			if (checkIfAlreadyRunning() == true)
			{
				cout << "TuxLCD is already running!" << endl;
				return EXIT_FAILURE;
			}
			
			/* Change the file mode mask */
			umask(0);
			
			/* Create a new SID for the child process */
			sid = setsid();
			if (sid < 0) 
			{
				/* Log the failure */
				cout << "ERROR: Creating SID failed" << endl;
				exit(EXIT_FAILURE);
			}
			
			/* Change the current working directory */
			if ((chdir("/")) < 0)
			{
				/* Log the failure */
				cout << "ERROR: Chdir failed" << endl;
				exit(EXIT_FAILURE);
			}
			
			TuxLCD *tuxlcd = new TuxLCD(configFile, foreground);
			tuxlcd->init();
			
			/* Close out the standard file descriptors */
			close(STDIN_FILENO);
			close(STDOUT_FILENO);
			close(STDERR_FILENO);
			
			tuxlcd->main();
			delete tuxlcd;
			remove(PID_FILE);
		}
	}
	else
	{
		if (checkIfAlreadyRunning() == true)
		{
			cout << "TuxLCD is already running (if I've made a mistake then delete " << PID_FILE << " and try again)!" << endl;
			return EXIT_FAILURE;
		}
		TuxLCD *tuxlcd = new TuxLCD(configFile, foreground);
		tuxlcd->init();
		tuxlcd->main();
		delete tuxlcd;
		remove(PID_FILE);
	}
	
	return EXIT_SUCCESS;
}
