#ifndef __TUXLCD_H__
#define __TUXLCD_H__

#include <vector>
#include "Driver.h"
#include "Plugin.h"

using namespace std;

#define VERSION			"0.1a"
#define DEFAULT_CONFIG		"/etc/tuxlcd/tuxlcd.conf"
#define PID_FILE		"/tmp/tuxlcd.pid"

#define LOG_LEVEL_DEBUG		5
#define LOG_LEVEL_INFO		4
#define LOG_LEVEL_WARN		3
#define LOG_LEVEL_ERROR		2
#define LOG_LEVEL_FATAL		1


struct PluginData
{
	create_plugin_t *create_plugin;
	Plugin *plugin_instance;
	destroy_plugin_t *destroy_plugin;
};


struct Screen
{
	char *name;
	unsigned char duration;
	bool enabled;
	char *customChars;
	char *line1;
	char *line2;
};



/**
 * The main TuxLCD class.
 *
 * \author David Bull
 * \version 0.1
 * \date Monday 7th August 2006
 */
class TuxLCD
{
	private:
		bool foreground;
		Driver *lcdDriver;
		create_driver_t *create_driver;
		destroy_driver_t *destroy_driver;
		unsigned char displayContrast;
		
		char *driverName;
		char *screenConfig;
		char *eventConfig;
		char *driverDir;
		char *pluginDir;
		char *customCharDir;
		
		pthread_t mainThread;
		vector<PluginData> plugins;
		vector<Screen> screens;
		size_t screenIndex;
		
		
		/**
		 *
		 * \param[in] contrast
		 */
		void transitionFadeIn(unsigned char contrast);
		
		
		/**
		 *
		 */
		void transitionFadeOut();
		
		
		/**
		 * Parses as line of text from a screen definition replacing all tokens with the values from the corresponding plugins.
		 * Custom character tokens are defined as [$CCx] where 'x' is 0-7.
		 * Plugin tokens are defined as [$name.property] where 'name' is the name of the plugin and 'property' is a property from the specified plugin.
		 * NOTE: If the buffer is not long enough to contain the processed output then the output is truncated.
		 * \param[in] line the line to parse.
		 * \param[in,out] buffer the buffer in which to place the processed output.
		 * \param[in] bufferLen the length of the output buffer.
		 * \return the number of characters placed in the output buffer.
		 */
		int parseScreenLine(char *line, char *buffer, size_t bufferLen);
		
		
		/**
		 * Loads the main TuxLCD configuration file.
		 * \param[in] configFile the full path to the configuration file.
		 * \throw TuxLCDException if the config file could not be opened or an option was missing.
		 */
		void loadConfig(const char *configFile);
		
		
		/**
		 * Loads the specified TuxLCD display driver.
		 * \param[in] fileName full path to driver file.
		 * \throw TuxLCDException if there was a problem loading the driver.
		 */
		void TuxLCD::loadDriver(const char *fileName);
		
		
		/**
		 * Loads the plugins.
		 * The plugins directory is scanned and all shared objects are loaded.
		 * An instance of each of the plugins is created and places in the plugins
		 * vector along with pointers to its create and destroy functions.
		 * \param[in] pluginDir the directory containing the plugins.
		 */
		void loadPlugins(const char *pluginDir);
		
		
		/**
		 * Unloads the plugins.
		 * Destroys the plugins and removes them from the plugins vector.
		 */
		void unloadPlugins();
		
		
		/**
		 * Loads the specified screen definition file.
		 * Screens are processed into the Screen structure and placed into the screens vector.
		 * \param[in] fileName full path to the screen definition file.
		 */
		void loadScreens(char *fileName);
		
		
		/**
		 * Unloads the screen definitions contained in the screens vector.
		 * The memory used by the screen structures are freed and removed from the screens vector.
		 */
		void unloadScreens();
		
		
		/**
		 * TODO
		 * \param[in] fileName
		 */
		void loadEvents(char *fileName);
		
		
		/**
		 * TODO
		 */
		void unloadEvents();
		
		
		/**
		 * Loads an LCDC Custom Character file from disk into multidimensional char array.
		 * The array must be char[8][8] and be allocated before being passed to this method.
		 * \param[in] filename full path to the custom character file.
		 * \param[in,out] customCharSet the char[8][8] array to load the custom characters into.
		 * \return 0 if the custom character file was loaded correctly, 1 if not.
		 */
		int loadCustomCharSet(const char *filename, char customCharSet[8][8]);
		
		
		/**
		 * Centers the given text within the given buffer by prepending and appending spaces.
		 * For example if we want to center "hello" into a buffer of size 9, the result will be "  hello  ".
		 * If the length of the text is larger then the buffer then the text is simply copied to the buffer and the extra characters are cuttoff.
		 * \param[in] text the text to center.
		 * \param[in] centerLength the space withn which to center the text.
		 * \param[in,out] centeredText the buffer that will contain the centered text.
		 */
		void TuxLCD::centerText(const char *text, size_t centerLength, char *centeredText);
	
	
	public:
		/**
		 * Constructor.
		 * \param[in] configFile full path to TuxLCD configuration file.
		 * \param[in] foreground if true then TuxLCD stays in the foreground, otherwise it runs as a deamon.
		 */
		TuxLCD(const char *configFile, bool foreground);
		
		
		/**
		 * Destrustor.
		 * Displays a goodbye message on the LCD, unloads the screens and plugins, and destroys the TuxLCD driver.
		 */
		~TuxLCD();
		
		
		/**
		 * Initialises TuxLCD.
		 * \return 0 on success and 1 on failure.
		 */
		int init();
		
		
		/**
		 * Main TuxLCD method.
		 * Initialises the TuxLCD Driver, loads the plugins and screen definitions,
		 * and then loops forever displaying the screens.
		 */
		void main();
		
		
		/**
		 * Prints the properties for the specified plugin to the console.
		 * If the specified plugin was not found a list of all available plugins is printed.
		 * \param[in] pluginName the name of the plugin.
		 */
		void printPluginProps(const char *pluginName);
		
		
		/**
		 * Called from a TuxLCD Driver when a key on the screens keypad is pressed.
		 * \param[in] key the key that was pressed.
		 */
		void lcdKeyPressed(char key);
		
		
		/**
		 * Logs a message to either the console or a seperate log file depending on whether we are running as a daemon or not.
		 * \param[in] level
		 * \param[in] msg
		 */
		void log(int level, string msg);
};


/**
 * Signal handler for SIGINT, SIGTERM, and SIGUSR1.
 * We trap SIGINT and SIGTERM so that we can perform a nice shutdown of TuxLCD.
 * SIGUSR1 is used to interrupt the sleep() between screens so that, for example, we can skip between screens.
 * \param[in] signum the signal number.
 */
extern "C" void tuxlcdSignalHandler(int signum);

#endif
