#include <cstdlib>		// Standard library definitions
#include <unistd.h>		// UNIX standard function definitions
#include <cstdio>		// Standard input/output definitions
#include <iostream>		// Classes for input and output operations
#include <string>		// Standard library to manipulate C strings
#include <pthread.h>		// POSIX threads
#include <csignal>		//

#include <fcntl.h>		// File control definitions
#include <termios.h>		// POSIX terminal control definitions

#include "Driver_LK202.h"
#include "../TuxLCD.h"
#include "../TuxLCDException.h"

using namespace std;


/**
 * Constructor.
 * \param[in] tuxlcd pointer back to tuxlcd for callbacks such as when the keypad on the LCD is pressed.
 * \throws TuxLCDException if the serial port could not be opened.
 */
Driver_LK202::Driver_LK202(TuxLCD *tuxlcd) :
	inputThreadRunning(false), backlightState(true), brightness(DEFAULT_BRIGHTNESS), contrast(DEFAULT_CONTRAST)
{
	tuxlcdPtr = tuxlcd;
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Driver_LK202: Constructing...\n");
	struct termios options;		// Serial port settings
	
	
	/*
	 * Open the serial port
	 */
	tuxlcdPtr->log(LOG_LEVEL_INFO, "Driver_LK202: Opening serial port...\n");
	fd = open(SERIAL_PORT, O_RDWR | O_NOCTTY | O_NDELAY);
	if (fd == -1)
	{
		//Could not open the port.
		string msg = "Driver_LK202: ERROR: Cannot open serial port: ";
		msg += SERIAL_PORT;
		msg += "\n";
		tuxlcdPtr->log(LOG_LEVEL_FATAL, msg);
		throw new TuxLCDException(msg.c_str());
	}
	fcntl(fd, F_SETFL, 0);
	
	
	/*
	 * Configure the serial port - 19200, 8N1, no handshaking
	 * http://www.easysw.com/~mike/serial/serial.html
	 */
	tuxlcdPtr->log(LOG_LEVEL_INFO, "Driver_LK202: Configuring serial port...\n");
	tcgetattr(fd, &options);
	options.c_cflag = B19200;
	options.c_cflag |= CS8;
	options.c_cflag |= CREAD;
	options.c_iflag = IGNPAR | IGNBRK;
	options.c_cflag |= CLOCAL;
	options.c_oflag = 0;
	options.c_lflag = 0;
	options.c_cc[VTIME] = 0;
	options.c_cc[VMIN] = 1;
	tcsetattr(fd, TCSANOW, &options);
	tcflush(fd, TCOFLUSH);
	tcflush(fd, TCIFLUSH);
	
	
	/*
	 * Initialise the LCD
	 */
	tuxlcdPtr->log(LOG_LEVEL_INFO, "Driver_LK202: Initialising LCD...\n");
	enableGpo(1, false);
	enableGpo(2, false);
	enableGpo(3, false);
	enableGpo(4, false);
	enableGpo(5, false);
	enableGpo(6, false);
	enableAutoLineWrap(false);
	enableAutoScroll(false);
	enableUnderlineCursor(false);
	enableBlinkingCursor(false);
	setBrightness(DEFAULT_BRIGHTNESS);
	setContrast(DEFAULT_CONTRAST);
	backlightOn(0);	
	
	int retval = pthread_create(&inputThread, NULL, monitorLcdKeypad, this);
	if (retval != 0)
		tuxlcdPtr->log(LOG_LEVEL_ERROR, "Driver_LK202: ERROR: Cannot create monitorLcdKeypad thread!\n");
}


/**
 * Destructor.
 */
Driver_LK202::~Driver_LK202()
{
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Driver_LK202: Destructing...\n");
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Driver_LK202: Waiting for monitorLcdKeypad thread to finish...\n");
	close(fd);
	inputThreadRunning = false;
	pthread_kill(inputThread, SIGUSR2);
	pthread_join(inputThread, NULL);
}


/**
 *
 * \param[in] lcdDriver
 */
void *monitorLcdKeypad(void *lcdDriver)
{
	//cout << "Driver_LK202: monitorLcdKeypad() entered\n";
	Driver_LK202 *driver = (Driver_LK202 *)lcdDriver;

	char buf[1];
	ssize_t bytesRead;
	driver->inputThreadRunning = true;
	
	signal(SIGUSR2, lcdKeypadThreadSignalHandler);
	while (driver->inputThreadRunning == true)
	{
		bytesRead = read(driver->fd, buf, 1);
		if (bytesRead == -1)
			break;
		driver->tuxlcdPtr->lcdKeyPressed(buf[0]);
	}

	//cout << "Driver_LK202: monitorLcdKeypad() finished\n";
}


/**
 *
 * \param[in] signum
 */
void lcdKeypadThreadSignalHandler(int signum)
{
	//cout << "Driver_LK202: lcdKeypadThreadSignalHandler() caught a signal" << endl;
}


/**
 *
 * \param[in] buffer
 * \param[in] length
 * \return
 */
int Driver_LK202::sendBuffer(char *buffer, int length)
{
	int n;
	n = write(fd, buffer, length);
	if (n < 0)
		return 1;
	else
		return 0;
}


/**
 * This command clears the display and resets the text insertion point to the top left of the screen.
 * \return
 */
int Driver_LK202::clearScreen()
{
	char buffer[] = {254, 88};
	return sendBuffer(buffer, 2);
}


/**
 * This command sets the display's contrast to [contrast], where [contrast] is a value between 0x00 and 0xFF (between 0 and 255).
 * Lower values cause 'on' elements in the display area to appear lighter, while higher values cause 'on' elements to appear darker.
 * \param[in] newContrast
 * \return
 */
int Driver_LK202::setContrast(unsigned char newContrast)
{
	int retval;
	char buffer[] = {254, 80, contrast};
	retval =  sendBuffer(buffer, 3);
	if (retval==0)
		contrast = newContrast;
	return retval;
}


/**
 * Returns the contrast of the display.
 * NOTE: There is n built-in way to retieve the contrast, instead we return the contrast value last passed to setContrast().
 * \return the contrast of the display.
 */
unsigned char Driver_LK202::getContrast()
{
	return contrast;
}


/**
 * This commands sets the display's backlight brightness to [brightness], where [brightness] is a value between 0x00 and 0xFF (between 0 and 255). 
 * Lower values will cause the backlight intensity to 'dim', while higher values cause the backlight's intensity to 'brighten'.
 * \param[in] newBrightness
 * \return
 */
int Driver_LK202::setBrightness(unsigned char newBrightness)
{
	int retval;
	char buffer[] = {254, 153, brightness};
	retval = sendBuffer(buffer, 3);
	if (retval==0)
		brightness = newBrightness;
	return retval;
}


/**
 * Returns the brightness of the display.
 * NOTE: There is n built-in way to retieve the brightness, instead we return the brightness value last passed to setBrightness().
 * \return the brightness of the display.
 */
unsigned char Driver_LK202::getBrightness()
{
	return brightness;
}


/**
 * This command turns on the backlight for a time of [minutes] minutes.
 * If [minutes] is zero (0), the backlight will remain on indefinitely.
 * \param[in] minutes
 * \return
 */
int Driver_LK202::backlightOn(char minutes)
{
	char buffer[] = {254, 66, minutes};
	backlightState = true;
	return sendBuffer(buffer, 3);
}


/**
 * This command turns the backlight off.
 * \return
 */
int Driver_LK202::backlightOff()
{
	char buffer[] = {254, 70};
	backlightState = false;
	return sendBuffer(buffer, 2);
}


/**
 * Toggles the backlight.
 * NOTE: There is no built-in toggle-backlight command, so we keep track of the state in the calls to backlightOn() and backlightOff().
 * Also because the display has an auto-off function, the backlight might go out and we won't know about.
 * This could be solved with threads and a timer, but that seems like to much effort at the moment.
 * \return
 */
int Driver_LK202::toggleBacklight()
{
	int retval;
	if (backlightState == true)
		retval = backlightOff();
	else
		retval = backlightOn(0);
	return retval;
}


/**
 * Returns the state of backlight.
 * NOTE: There is no built-in way to retrieve the backlight state, so we keep track of the state in the calls to backlightOn() and backlightOff().
 * Also because the display has an auto-off function, the backlight might go out and we won't know about.
 * This could be solved with threads and a timer, but that seems like to much effort at the moment.
 * \return the state of backlight.
 */
bool Driver_LK202::getBacklightState()
{
	return backlightState;
}


/**
 * This command turns ON or OFF any of the general purpose outputs. [gpo #] is 1 to 6. 
 * Note that ON means that the output is pulled low (ground via 240ohms).
 * Note that OFF means that the output is high.
 * \param[in] gpo 1-6
 * \param enabled
 * \return
 */
int Driver_LK202::enableGpo(char gpo, bool enabled)
{
	// 87 = on, 86 = off
	char buffer[] = {254, (enabled ? 87 : 86), gpo};
	if (gpo<1 || gpo>6)
		return 1;
	return sendBuffer(buffer, 3);
}


/**
 * Enables or diables automatic line wrapping. 
 * When disabled, characters beyond the end of a line will be lost.
 * Note that this is not 'word wrapping' and wraps may occur in the middle of a word.
 * \param enabled
 * \return
 */
int Driver_LK202::enableAutoLineWrap(bool enabled)
{
	// 67 = on, 68 = off
	char buffer[] = {254, (enabled ? 67 : 68)};
	return sendBuffer(buffer, 2);
}


/**
 * Enables or diables automatic scrolling
 * When auto scrolling is on, it causes the display to shift the entire display's contents up to make room for a new line of text when the text reaches the scroll position (the bottom right character position).
 * When auto scrolling is disabled, text will wrap to the top left corner of the display area.
 * Existing text in the display area is not erased before new text is placed. A series of 'spaces' followed by a 'Cursor home' command may be used to erase the top line of text.
 * \param enabled
 * \return
 */
int Driver_LK202::enableAutoScroll(bool enabled)
{
	// 81 = on, 82 = off
	char buffer[] = {254, (enabled ? 81 : 82)};
	return sendBuffer(buffer, 2);
}


/**
 * This command sets the cursor position (text insertion point) to the [column] and [row] specified.
 * Columns have values from 1 to 20 (0x01 to 0x14) and rows have values of 1 and 2 (0x01 and 0x02).
 * \param[in] column 1-20
 * \param[in] row 1-2
 * \return
 */
int Driver_LK202::setCursorPosition(char column, char row)
{
	char buffer[] = {254, 71, column, row};
	if ((column<1 || column>20) || (row<1 || row>2))
		return 1;
	return sendBuffer(buffer, 4);
}


/**
 * This command moves the cursor position (text insertion point) to the top left of the display area.
 * \return
 */
int Driver_LK202::sendCursorHome()
{
	char buffer[] = {254, 72};
	return sendBuffer(buffer, 2);
}


/**
 * Turns on or off the underline cursor.
 * The cursor shows the current text insertion point.
 * Both underline and blinking cursors may be turned on or off independently.
 * \param enabled
 * \return
 */
int Driver_LK202::enableUnderlineCursor(bool enabled)
{
	// 74 = on, 75 = off
	char buffer[] = {254, (enabled ? 74 : 75)};
	return sendBuffer(buffer, 2);
}


/**
 * Turns on or off the blinking block cursor.
 * The cursor shows the current text insertion point.
 * Both blinking and underline cursors may be turned on or off independently.
 * \param[in] enabled
 * \return
 */
int Driver_LK202::enableBlinkingCursor(bool enabled)
{
	// 83 = on, 84 = off
	char buffer[] = {254, (enabled ? 83 : 84)};
	return sendBuffer(buffer, 2);
}


/**
 * Moves the cursor one position to the left but does not erase any character that may be in that position.
 * Note that this command moves the text insertion point even if the cursor is turned off.
 * A 'destructive backspace', which erases the character to the left of the original position, may be done by issuing the following sequence: cursor left, space, cursor left.
 * \return
 */
int Driver_LK202::cursorLeft()
{
	char buffer[] = {254, 76};
	return sendBuffer(buffer, 2);
}


/**
 * Moves the cursor one position to the right but does not erase any character that may be in that position.
 * Note that this command moves the text insertion point even if the cursor is turned off.
 * \return
 */
int Driver_LK202::cursorRight()
{
	char buffer[] = {254, 77};
	return sendBuffer(buffer, 2);
}


/**
 * Defines a custom character.
 * \param[in] charNum 0-7
 * \param[in] charData 8 byte character map
 * \return
 */
int Driver_LK202::defineCustomChar(char charNum, const char charData[8])
{
	char buffer[11];
	if (charNum<0 && charNum>7)
		return 1;
	buffer[0] = 254;
	buffer[1] = 78;
	buffer[2] = charNum;
	memcpy(buffer+3, charData, 8);
	//cout << "Driver_LK202: Uploading custom char %i: [%i, %i, %i, %i, %i, %i, %i, %i]\n", buffer[2], buffer[3], buffer[4], buffer[5], buffer[6], buffer[7], buffer[8], buffer[9], buffer[10]);
	return sendBuffer(buffer, 11);
}


/**
 * Defines 8 custom characters.
 * \param[in] customCharSet
 * \return
 */
int Driver_LK202::defineCustomCharSet(const char customCharSet[8][8])
{
	int n, retVal;
	for (n=0; n<=7; n++)
	{
		retVal = defineCustomChar(n, customCharSet[n]);
		if (retVal != 0)
			return retVal;
	}
	return 0;
}

