#include <pthread.h>
#include "../TuxLCD.h"
#include "../Driver.h"
using namespace std;

#ifndef __DRIVER_LK202_H__
#define __DRIVER_LK202_H__

#define SERIAL_PORT		"/dev/ttyS0"
#define BAUD_RATE		B19200

#define DEFAULT_CONTRAST	128
#define DEFAULT_BRIGHTNESS	128

#define SCREEN_COLS		20
#define SCREEN_ROWS		2
#define SCREEN_GPOS		6


/**
 * TuxLCD display driver for Matrix Orbital LK202-25.
 *
 * \author David Bull
 * \version 0.1
 * \date Monday 7th August 2006
 */
class Driver_LK202 : public Driver
{
	private:
		TuxLCD *tuxlcdPtr;
		pthread_t inputThread;
		bool inputThreadRunning;
		int fd;
		
		bool backlightState;
		unsigned char brightness;
		unsigned char contrast;
		
		friend void *monitorLcdKeypad(void *lcdDriver);
		friend void lcdKeypadThreadSignalHandler(int signum);
	
	
	public:
		/**
		 * Constructor.
		 * \param[in] tuxlcd pointer back to tuxlcd for callbacks such as when the keypad on the LCD is pressed.
		 * \throws TuxLCDException if the serial port could not be opened.
		 */
		Driver_LK202(TuxLCD *tuxlcd);
		
		
		/**
		 * Destrustor.
		 */
		~Driver_LK202();
		
		
		/**
		 * Returns the number of columns the screen has.
		 * \return the number of columns the screen has.
		 */
		int getScreenColumns()
		{
			return SCREEN_COLS;
		}
		
		
		/**
		 * Returns the number of rows the screen has.
		 * \return the number of rows the screen has.
		 */
		int getScreenRows()
		{
			return SCREEN_ROWS;
		}
		
		
		/**
		 * Returns the number of GPOs the screen has.
		 * \return the number of GPOs the screen has.
		 */
		int getScreenGPOs()
		{
			return SCREEN_GPOS;
		}
		
		
		/**
		 *
		 * \param[in] buffer
		 * \param[in] length
		 * \return
		 */
		int sendBuffer(char *buffer, int length);
		
		
		/**
		 * This command clears the display and resets the text insertion point to the top left of the screen.
		 * \return
		 */
		int clearScreen();
		
		
		/**
		 * This command sets the display's contrast to [contrast], where [contrast] is a value between 0x00 and 0xFF (between 0 and 255).
		 * Lower values cause 'on' elements in the display area to appear lighter, while higher values cause 'on' elements to appear darker.
		 * \param[in] newContrast
		 * \return
		 */
		int setContrast(unsigned char newContrast);
		
		
		/**
		 * Returns the contrast of the display.
		 * NOTE: There is n built-in way to retieve the contrast, instead we return the contrast value last passed to setContrast().
		 * \return the contrast of the display.
		 */
		unsigned char getContrast();
		
		
		/**
		 * This commands sets the display's backlight brightness to [brightness], where [brightness] is a value between 0x00 and 0xFF (between 0 and 255). 
		 * Lower values will cause the backlight intensity to 'dim', while higher values cause the backlight's intensity to 'brighten'.
		 * \param[in] newBrightness
		 * \return
		 */
		int setBrightness(unsigned char newBrightness);
		
		
		/**
		 * Returns the brightness of the display.
		 * NOTE: There is n built-in way to retieve the brightness, instead we return the brightness value last passed to setBrightness().
		 * \return the brightness of the display.
		 */
		unsigned char getBrightness();
		
		
		/**
		 * This command turns on the backlight for a time of [minutes] minutes.
		 * If [minutes] is zero (0), the backlight will remain on indefinitely.
		 * \param[in] minutes
		 * \return
		 */
		int backlightOn(char minutes);
		
		
		/**
		 * This command turns the backlight off.
		 * \return
		 */
		int backlightOff();
		
		
		/**
		 * Toggles the backlight.
		 * NOTE: There is no built-in toggle-backlight command, so we keep track of the state in the calls to backlightOn() and backlightOff().
		 * Also because the display has an auto-off function, the backlight might go out and we won't know about.
		 * This could be solved with threads and a timer, but that seems like to much effort at the moment.
		 * \return
		 */
		int toggleBacklight();
		
		
		/**
		 * Returns the state of backlight.
		 * NOTE: There is no built-in way to retrieve the backlight state, so we keep track of the state in the calls to backlightOn() and backlightOff().
		 * Also because the display has an auto-off function, the backlight might go out and we won't know about.
		 * This could be solved with threads and a timer, but that seems like to much effort at the moment.
		 * \return the state of backlight.
		 */
		bool getBacklightState();
		
		
		/**
		 * This command turns ON or OFF any of the general purpose outputs. [gpo #] is 1 to 6. 
		 * Note that ON means that the output is pulled low (ground via 240ohms).
		 * Note that OFF means that the output is high.
		 * \param[in] gpo 1-6
		 * \param[in] enabled
		 * \return
		 */
		int Driver_LK202::enableGpo(char gpo, bool enabled);
		
		
		/**
		 * Enables or diables automatic line wrapping. 
		 * When disabled, characters beyond the end of a line will be lost.
		 * Note that this is not 'word wrapping' and wraps may occur in the middle of a word.
		 * \param[in] enabled
		 * \return
		 */
		int enableAutoLineWrap(bool enabled);
		
		
		/**
		 * Enables or disables automatic scrolling
		 * When auto scrolling is on, it causes the display to shift the entire display's contents up to make room for a new line of text when the text reaches the scroll position (the bottom right character position).
		 * When auto scrolling is disabled, text will wrap to the top left corner of the display area.
		 * Existing text in the display area is not erased before new text is placed. A series of 'spaces' followed by a 'Cursor home' command may be used to erase the top line of text.
		 * \param[in] enabled
		 * \return
		 */
		int enableAutoScroll(bool enabled);
		
		
		/**
		 * This command sets the cursor position (text insertion point) to the [column] and [row] specified.
		 * Columns have values from 1 to 20 (0x01 to 0x14) and rows have values of 1 and 2 (0x01 and 0x02).
		 * \param[in] column 1-20
		 * \param[in] row 1-2
		 * \return
		 */
		int setCursorPosition(char column, char row);
		
		
		/**
		 * This command moves the cursor position (text insertion point) to the top left of the display area.
		 * \return
		 */
		int sendCursorHome();
		
		
		/**
		 * Turns on or off the underline cursor.
		 * The cursor shows the current text insertion point.
		 * Both underline and blinking cursors may be turned on or off independently.
		 * \param[in] enabled
		 * \return
		 */
		int enableUnderlineCursor(bool enabled);
		
		
		/**
		 * Turns on or off the blinking block cursor.
		 * The cursor shows the current text insertion point.
		 * Both blinking and underline cursors may be turned on or off independently.
		 * \param[in] enabled
		 * \return
		 */
		int enableBlinkingCursor(bool enabled);
		
		
		/**
		 * Moves the cursor one position to the left but does not erase any character that may be in that position.
		 * Note that this command moves the text insertion point even if the cursor is turned off.
		 * A 'destructive backspace', which erases the character to the left of the original position, may be done by issuing the following sequence: cursor left, space, cursor left.
		 * \return
		 */
		int cursorLeft();
		
		
		/**
		 * Moves the cursor one position to the right but does not erase any character that may be in that position.
		 * Note that this command moves the text insertion point even if the cursor is turned off.
		 * \return
		 */
		int cursorRight();
		
		
		/**
		 * Defines a custom characters.
		 * \param[in] charNum 0-7
		 * \param[in] charCata 8 byte character map
		 * \return
		 */
		int defineCustomChar(char charNum, const char charCata[8]);
		
		
		/**
		 * Defines 8 custom characters.
		 * \param[in] customCharSet
		 * \return
		 */
		int defineCustomCharSet(const char customCharSet[8][8]);
};


/**
 *
 * \param[in] lcdDriver
 */
extern void *monitorLcdKeypad(void *lcdDriver);


/**
 *
 * \param[in] signum
 */
extern void lcdKeypadThreadSignalHandler(int signum);



// the class factories

extern "C" Driver* create(void *tuxlcd)
{
	return new Driver_LK202((TuxLCD*)tuxlcd);
}

extern "C" void destroy(Driver *d)
{
	delete d;
}


#endif
