#include <cstdlib>		// Standard library definitions
#include <unistd.h>		// UNIX standard function definitions
#include <cstdio>		// Standard input/output definitions
#include <cmath>		// Standard library for mathematic operations
#include <iostream>		// Classes for input and output operations
#include <string>		// Standard library to manipulate C strings
#include <vector>		//

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "sensors.h" 
#include "error.h"
#include "chips.h"

#include <glib.h>

#include "Plugin_Sensors.h"
#include "../TuxLCD.h"
#include "../TuxLCDException.h"

using namespace std;


/**
 * Constructor.
 * Attempts to initialise libsensors. If an error occurs then libsensors data will not be available.
 * Data will still be availble from hddtemp and nvidia-settings (assuming they both work).
 * \param[in] tuxlcd pointer back to tuxlcd for callbacks such as when the keypad on the LCD is pressed.
 */
Plugin_Sensors::Plugin_Sensors(TuxLCD *tuxlcd) :
	sensorsInitialised(false), hdd_str(NULL), hdd_str_size(0)
{
	tuxlcdPtr = tuxlcd;
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Plugin_Sensors: Constructing...\n");
	
	/*
	 * Open sensors config file and initialise libsensors
	 */
	int retval;
	FILE *config_file = fopen(DEFAULT_SENSORS_CONFIG_FILE, "r");
	if (!config_file)
	{
		string msg = "Plugin_Sensors: ERROR: Could not open sensors config file '";
		msg.append(DEFAULT_SENSORS_CONFIG_FILE);
		msg.append("'\n");
		tuxlcdPtr->log(LOG_LEVEL_ERROR, msg);
		tuxlcdPtr->log(LOG_LEVEL_WARN, "Plugin_Sensors: WARNING: Plugin_Sensors: libsensors data will not be available\n");
	}
	retval = sensors_init(config_file);
	if (retval != 0)
	{
		string msg = "Plugin_Sensors: ERROR: ";
		msg.append(sensors_strerror(retval));
		msg.append("\n");
		tuxlcdPtr->log(LOG_LEVEL_ERROR, msg);
		tuxlcdPtr->log(LOG_LEVEL_WARN, "Plugin_Sensors: WARNING: libsensors data will not be available\n");
	}
	fclose(config_file);
	sensorsInitialised = true;
	
	
	/*
	 * Check if hddtemp daemon is available
	 */
	char *reply = queryHddtemp(DEFAULT_HDDTEMP_SERVER, DEFAULT_HDDTEMP_PORT);
	if (reply == NULL)
	{
		tuxlcdPtr->log(LOG_LEVEL_ERROR, "Plugin_Sensors: ERROR: Could not connect to hddtemp daemon\n");
		tuxlcdPtr->log(LOG_LEVEL_WARN, "Plugin_Sensors: WARNING: hddtemp data will not be available\n");
	}
	free(reply);
	
	
	/*
	 * Check if nvidia-settings can report GPU temperature
	 */
	if (getNvidiaGpuTemp() == 0.0)
	{
		tuxlcdPtr->log(LOG_LEVEL_ERROR, "Plugin_Sensors: ERROR: Could not get nvidia GPU core temperature\n");
		tuxlcdPtr->log(LOG_LEVEL_WARN, "Plugin_Sensors: WARNING: nvidia GPU core temperature data will not be available\n");
	}
}


/**
 * Destrustor.
 */
Plugin_Sensors::~Plugin_Sensors()
{
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Plugin_Sensors: Destructing...\n");
	free(hdd_str);
}


/**
 * Returns a list of properties available from this plugin.
 * Property values can then be retrieved by calling getProperty(propertyName).
 * \see getProperty(propertyName).
 * \return a list of properties available from this plugin.
 */
vector<string> Plugin_Sensors::getProperties()
{
	unsigned int i;
	vector<string> props;
	props.reserve(10);
	
	vector<string> sensorFeatures = getSensorFeatures();
	for (i=0; i<sensorFeatures.size(); i++)
		props.push_back(PROPERTY_SENSOR + sensorFeatures[i]);
		
	vector<string> drives = getDrives();
	for (i=0; i<drives.size(); i++)
		props.push_back(PROPERTY_HDDTEMP + drives[i]);
		
	props.push_back(PROPERTY_NVIDIATEMP);
	
	return props;
}


/**
 * Returns the value of the specified property, or NULL ifthe propertyName is not valid.
 * A list of valid property names is available from getProperties().
 * \see getProperties()
 * \param[in] propertyName the name of the property.
 * \return the value of the specified property, or NULL if the propertyName is not valid.
 */
char* Plugin_Sensors::getProperty(const char *propertyName)
{
	if (strlen(propertyName)>strlen(PROPERTY_SENSOR) && strncmp(propertyName,PROPERTY_SENSOR,strlen(PROPERTY_SENSOR))==0)
	{
		char *buf = (char *)calloc(7, sizeof(char));
		double sensor = getSensorFeature(propertyName+strlen(PROPERTY_SENSOR));
		//sprintf(buf, "%.1f", sensor);
		sprintf(buf, "%.0f", round(sensor));
		return buf;
	}
	
	else if (strlen(propertyName)>strlen(PROPERTY_HDDTEMP) && strncmp(propertyName,PROPERTY_HDDTEMP,strlen(PROPERTY_HDDTEMP))==0)
	{
		char *buf = (char *)calloc(3, sizeof(char));
		int temp = getDriveTemp(propertyName+strlen(PROPERTY_HDDTEMP));
		//sprintf(buf, "%i", temp);
		sprintf(buf, "%.0f", round(temp));
		return buf;
	}
	
	else if (strcmp(propertyName, PROPERTY_NVIDIATEMP) == 0)
	{
		char *buf = (char *)calloc(6, sizeof(char));
		float temp = getNvidiaGpuTemp();
		//sprintf(buf, "%.1f", temp);
		sprintf(buf, "%.0f", round(temp));
		return buf;
	}
	
	else
	{
		string msg = "Plugin_SysInfo: Invalid property '";
		msg.append(propertyName);
		msg.append("'\n");
		tuxlcdPtr->log(LOG_LEVEL_WARN, msg);
		return NULL;
	}
}


/**
 * Outputs all properties and their values to stdout.
 * Used for testing and debugging.
 */
void Plugin_Sensors::dumpProperties()
{
	cout << "Plugin_Sensors: Dumping properties...\n";
	vector<string> props = getProperties();
	for (size_t i=0; i<props.size(); i++)
	{
		cout << "  " << props[i] << "\t\t\t" << getProperty(props[i].c_str()) << "\n";
	}
}


/**
 * Returns a list of detected sensor features.
 * \return a list of detected sensor features.
 */
vector<string> Plugin_Sensors::getSensorFeatures()
{
	int nr = 0;
	const sensors_chip_name *chip;
	vector<string> sensorFeatures;
	if (sensorsInitialised == false)
		return sensorFeatures;
	sensorFeatures.reserve(10);
	while (1)
	{
		chip = sensors_get_detected_chips(&nr);
		if (chip == NULL)
			break;
		
		int a = 0, b = 0;
		const sensors_feature_data *data;
		while(1)
		{
			data = sensors_get_all_features(*chip, &a, &b);
			if (data == NULL)
				break;
			
			if (strstr(data->name, "_min")==NULL && strstr(data->name, "_div")==NULL && strstr(data->name, "_over")==NULL && strstr(data->name, "_hyst")==NULL)
				sensorFeatures.push_back(data->name);
		}
	}
	return sensorFeatures;
}


/**
 * Returns the value of the specified feature.
 * \param[in] featureName the name of the feature (e.g. "temp1").
 * \return the value of the specified feature.
 */
double Plugin_Sensors::getSensorFeature(const char* featureName)
{
	int retval;
	int nr = 0;
	const sensors_chip_name *chip;
	if (sensorsInitialised == false)
		return 0.0;
	while (1)
	{
		chip = sensors_get_detected_chips(&nr);
		if (chip == NULL)
			break;	
		
		int a = 0, b = 0;
		double reading;
		const sensors_feature_data *data;
		while(1)
		{
			data = sensors_get_all_features(*chip, &a, &b);
			if (data == NULL)
				break;
			
			if (strcmp(data->name, featureName)==0)
			{
				retval = sensors_get_feature(*chip, data->number, &reading);
				if (retval == 0)
					return reading;
				else
					return 0.0;
			}
		}
	}
	return 0.0;
}


/**
 *
 * \param[in] server
 * \param[in] port
 * \return
 */
char* Plugin_Sensors::queryHddtemp(const char *server, unsigned int port)
{
	struct sockaddr_in serv_addr;
	struct hostent *host;
	int sockfd;
	unsigned int n, pos;
	char buff[2];
	
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0)
		return NULL;	
	host = gethostbyname(server);
	if (host == NULL)
		return NULL;
	memcpy(&serv_addr.sin_addr.s_addr, host->h_addr, host->h_length);
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);
	if (connect(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0)
		return NULL;
	
	if(hdd_str == NULL) 
	{
		hdd_str_size = 2;
		hdd_str = (char *)malloc(hdd_str_size * sizeof(char));
		if (hdd_str == NULL)
			return NULL;
	}
	pos = 0;
	while ((n = read(sockfd, &buff, sizeof(buff))) > 0)
	{
		if (pos+n+1 > hdd_str_size)
		{
			hdd_str_size *= 2;
			hdd_str = (char *)realloc(hdd_str, hdd_str_size);
			if (hdd_str == NULL)
				return NULL;
		}
		strncpy(hdd_str+pos, buff, n);
		pos += n;
	}
	hdd_str[pos] = 0;
	
	close(sockfd);
	return strdup(hdd_str);
}


/**
 * Returns a list of drives reported from hddtemp.
 * \return a list of drives reported from hddtemp.
 */
vector<string> Plugin_Sensors::getDrives()
{
	char *reply = queryHddtemp(DEFAULT_HDDTEMP_SERVER, DEFAULT_HDDTEMP_PORT);
	vector<string> drives;
	if (reply == NULL)
		return drives;
	char delim = reply[0];
	char *replyPtr = reply;
	char *nextDelim = reply;
	char *buf = NULL;
	
	while (*(nextDelim+1) != 0)
	{
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
		buf = (char *)realloc(buf, nextDelim-replyPtr+1);
		strncpy(buf, replyPtr, nextDelim-replyPtr);
		buf[nextDelim-replyPtr] = 0;
		drives.push_back(buf);
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
	}
	
	free(buf);
	free(reply);
	return drives;
}


/**
 * Returns the temperature of the specifed drive as reported by hddtemp.
 * \param[in] drive the drive.
 * \return the temperature of the drive.
 */
short Plugin_Sensors::getDriveTemp(const char *drive)
{
	char *reply = queryHddtemp(DEFAULT_HDDTEMP_SERVER, DEFAULT_HDDTEMP_PORT);
	if (reply == NULL)
		return 0;
	char delim = reply[0];
	char *replyPtr = reply;
	char *nextDelim = reply;
	char *buf = NULL;
	bool driveFound = false;
	short temp = 0;
	
	while (*(nextDelim+1) != 0)
	{
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
		if ((strlen(drive) <= nextDelim-replyPtr) && (strncmp(replyPtr, drive, strlen(drive)) == 0))
			driveFound = true;
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
		if (driveFound == true)
		{
			buf = (char *)calloc(nextDelim-replyPtr+1, sizeof(char));
			strncpy(buf, replyPtr, nextDelim-replyPtr);
			temp = atoi(buf);
			break;
		}
		replyPtr = nextDelim+1;
		nextDelim = strchr(replyPtr, delim);
	}
	
	free(buf);
	free(reply);
	return temp;
}


/**
 * Returns the nvidia GPU core temperature using the nvidia-settings command.
 * \return the nvidia GPU core temperature.
 */
float Plugin_Sensors::getNvidiaGpuTemp()
{
	char *args[] = { "nvidia-settings", "-q", "GPUCoreTemp", NULL };
	char *output = NULL;
	gboolean result;
	gfloat temp = 0.0f;

	result = g_spawn_sync(NULL, args, NULL, (GSpawnFlags)(G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL), NULL, NULL, &output, NULL, NULL, NULL);
	if (result && output)
		result = (sscanf(output, " Attribute %*s %*s %f", &temp) == 1);
	g_free(output);

	return temp;
}
