#include <string>		// String function definitions
#include <vector>		//
#include "../TuxLCD.h"
#include "../Plugin.h"
using namespace std;


#define DEFAULT_SENSORS_CONFIG_FILE	"/etc/sensors.conf"
#define DEFAULT_HDDTEMP_SERVER		"127.0.0.1"
#define DEFAULT_HDDTEMP_PORT		7634


#define PROPERTY_SENSOR		"SENSOR_"
#define PROPERTY_HDDTEMP	"HDDTEMP_"
#define PROPERTY_NVIDIATEMP	"NVIDIATEMP"



/**
 * Sensor information plugin for TuxLCD.
 *
 * \author David Bull
 * \version 0.1
 * \date Tuesday 8th August 2006
 */
class Plugin_Sensors : public Plugin
{
	private:
		TuxLCD *tuxlcdPtr;
		bool sensorsInitialised;
		char *hdd_str;
		unsigned int hdd_str_size;
		
	
		/**
		 * Returns a list of detected sensor features.
		 * \return a list of detected sensor features.
		 */
		vector<string> getSensorFeatures();
		
		
		/**
		 * Returns the value of the specified feature.
		 * \param[in] featureName the name of the feature (e.g. "temp1").
		 * \return the value of the specified feature.
		 */
		double getSensorFeature(const char* featureName);
		
		
		/**
		 *
		 * \param[in] server
		 * \param[in] port
		 * \return
		 */
		char* queryHddtemp(const char *server, unsigned int port);
		
		
		/**
		 * Returns a list of drives reported from hddtemp.
		 * \return a list of drives reported from hddtemp.
		 */
		vector<string> getDrives();
		
		/**
		 * Returns the temperature of the specifed drive as reported by hddtemp.
		 * \param[in] drive the drive.
		 * \return the temperature of the drive.
		 */
		short getDriveTemp(const char *drive);
		
		
		/**
		 * Returns the nvidia GPU core temperature using the nvidia-settings command.
		 * \return the nvidia GPU core temperature.
		 */
		float Plugin_Sensors::getNvidiaGpuTemp();
	
	
	public:
		/**
		 * Constructor.
		 * Attempts to initialise libsensors. If an error occurs then libsensors data will not be available.
 		 * Data will still be availble from hddtemp and nvidia-settings (assuming they both work).
		 * \param[in] tuxlcd pointer back to tuxlcd for callbacks such as when the keypad on the LCD is pressed.
		 */
		Plugin_Sensors(TuxLCD *tuxlcd);
		
		
		/**
		 * Destrustor.
		 */
		~Plugin_Sensors();
		
		
		/**
		 * Returns the name of this plugin.
		 * \return the plugin name.
		 */
		const char* getPluginName()
		{
			return "Sensors";
		}
		
		
		/**
		 * Returns a list of properties available from this plugin.
		 * Property values can then be retrieved by calling getProperty(propertyName).
		 * \see getProperty(propertyName).
		 * \return a list of properties available from this plugin.
		 */
		vector<string> getProperties();
		
		
		/**
		 * Returns the value of the specified property, or NULL ifthe propertyName is not valid.
		 * A list of valid property names is available from getProperties().
		 * \see getProperties()
		 * \param[in] propertyName the name of the property.
		 * \return the value of the specified property, or NULL ifthe propertyName is not valid.
		 */
		char* getProperty(const char *propertyName);
		
		
		/**
		 * Outputs all properties and their values to stdout.
		 * Used for testing and debugging.
		 */
		void dumpProperties();
};


// the class factories

extern "C" Plugin* create(void *tuxlcd)
{
	return new Plugin_Sensors((TuxLCD*)tuxlcd);
}

extern "C" void destroy(Plugin *p)
{
	delete p;
}
