#include <cstdlib>		// Standard library definitions
#include <unistd.h>		// UNIX standard function definitions
#include <cstdio>		// Standard input/output definitions
#include <cmath>		// Standard library for mathematic operations
#include <iostream>		// Classes for input and output operations
#include <string>		// Standard library to manipulate C strings
#include <vector>		//

#include <sys/utsname.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <glibtop.h>
#include <glibtop/cpu.h>
#include <glibtop/mem.h>
#include <glibtop/fsusage.h>
#include <glibtop/mountlist.h>
#include <glibtop/netlist.h>
#include <glibtop/netload.h>

#include "Plugin_SysInfo.h"
#include "../TuxLCD.h"
#include "../TuxLCDException.h"

using namespace std;


/**
 * Constructor.
 * \param[in] tuxlcd pointer back to tuxlcd for callbacks such as when the keypad on the LCD is pressed.
 */
Plugin_SysInfo::Plugin_SysInfo(TuxLCD *tuxlcd)
{
	tuxlcdPtr = tuxlcd;
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Plugin_SysInfo: Constructing...\n");
	uname(&name);
	//glibtop_init();
}


/**
 * Destrustor.
 */
Plugin_SysInfo::~Plugin_SysInfo()
{
	tuxlcdPtr->log(LOG_LEVEL_DEBUG, "Plugin_SysInfo: Destructing...\n");
	glibtop_close();
}


/**
 * Returns a list of properties available from this plugin.
 * Property values can then be retrieved by calling getProperty(propertyName).
 * \see getProperty(propertyName).
 * \return a list of properties available from this plugin.
 */
vector<string> Plugin_SysInfo::getProperties()
{
	unsigned int i;
	vector<string> props;
	props.reserve(50);
	
	props.push_back(PROPERTY_TIME);
	props.push_back(PROPERTY_DATE);
	
	props.push_back(PROPERTY_SYSNAME);
	props.push_back(PROPERTY_NODENAME);
	props.push_back(PROPERTY_RELEASE);
	props.push_back(PROPERTY_VERSION);
	props.push_back(PROPERTY_MACHINE);
	
	props.push_back(PROPERTY_CPU_LOAD);
	props.push_back(PROPERTY_MEMORY_TOTAL);
	props.push_back(PROPERTY_MEMORY_USED);
	props.push_back(PROPERTY_MEMORY_FREE);
	props.push_back(PROPERTY_MEMORY_LOAD);
	
	vector<string> fileSystems = getFilesystems();
	for (i=0; i<fileSystems.size(); i++)
		props.push_back(PROPERTY_DISK_TOTAL + fileSystems[i]);
	for (i=0; i<fileSystems.size(); i++)
		props.push_back(PROPERTY_DISK_USED + fileSystems[i]);
	for (i=0; i<fileSystems.size(); i++)
		props.push_back(PROPERTY_DISK_FREE + fileSystems[i]);
	for (i=0; i<fileSystems.size(); i++)
		props.push_back(PROPERTY_DISK_LOAD + fileSystems[i]);
	
	vector<string> netInterfaces = getNetworkInterfaces();
	for (i=0; i<netInterfaces.size(); i++)
		props.push_back(PROPERTY_NET_ADDR + netInterfaces[i]);
	for (i=0; i<netInterfaces.size(); i++)
		props.push_back(PROPERTY_NET_DATA_IN + netInterfaces[i]);
	for (i=0; i<netInterfaces.size(); i++)
		props.push_back(PROPERTY_NET_DATA_OUT + netInterfaces[i]);
	for (i=0; i<netInterfaces.size(); i++)
		props.push_back(PROPERTY_NET_SPEED_IN + netInterfaces[i]);
	for (i=0; i<netInterfaces.size(); i++)
		props.push_back(PROPERTY_NET_SPEED_OUT + netInterfaces[i]);
	
	return props;
}


/**
 * Returns the value of the specified property, or NULL if the propertyName is not valid.
 * A list of valid property names is available from getProperties().
 * \see getProperties()
 * \param[in] propertyName the name of the property.
 * \return the value of the specified property, or NULL ifthe propertyName is not valid.
 */
char* Plugin_SysInfo::getProperty(const char *propertyName)
{
	if (strcmp(propertyName, PROPERTY_TIME) == 0)
		return getTime();
	else if (strcmp(propertyName, PROPERTY_DATE) == 0)
		return getDate();

	else if (strcmp(propertyName, PROPERTY_SYSNAME) == 0)
		return strdup(name.sysname);
	else if (strcmp(propertyName, PROPERTY_NODENAME) == 0)
		return strdup(name.nodename);
	else if (strcmp(propertyName, PROPERTY_RELEASE) == 0)
		return strdup(name.release);
	else if (strcmp(propertyName, PROPERTY_VERSION) == 0)
		return strdup(name.version);
	else if (strcmp(propertyName, PROPERTY_MACHINE) == 0)
		return strdup(name.machine);
	
	else if (strcmp(propertyName, PROPERTY_CPU_LOAD) == 0)
	{
		char *buf = (char *)calloc(5, sizeof(char));
		short cpuLoad = getCpuLoad();
		sprintf(buf, "%i", cpuLoad);
		return buf;
	}
	else if (strcmp(propertyName, PROPERTY_MEMORY_TOTAL) == 0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		short memTotal = getMemoryTotal();
		sprintf(buf, "%i", memTotal);
		return buf;
	}
	else if (strcmp(propertyName, PROPERTY_MEMORY_USED) == 0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		short memUsed = getMemoryUsed();
		sprintf(buf, "%i", memUsed);
		return buf;
	}
	else if (strcmp(propertyName, PROPERTY_MEMORY_FREE) == 0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		short memFree = getMemoryFree();
		sprintf(buf, "%i", memFree);
		return buf;
	}
	else if (strcmp(propertyName, PROPERTY_MEMORY_LOAD) == 0)
	{
		char *buf = (char *)calloc(5, sizeof(char));
		short memLoad = getMemoryLoad();
		sprintf(buf, "%i", memLoad);
		return buf;
	}
	
	else if (strlen(propertyName)>strlen(PROPERTY_DISK_TOTAL) && strncmp(propertyName,PROPERTY_DISK_TOTAL,strlen(PROPERTY_DISK_TOTAL))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int diskSize = getFilesystemTotal(propertyName+strlen(PROPERTY_DISK_TOTAL));
		sprintf(buf, "%i", diskSize);
		return buf;
	}
	else if (strlen(propertyName)>strlen(PROPERTY_DISK_USED) && strncmp(propertyName,PROPERTY_DISK_USED,strlen(PROPERTY_DISK_USED))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int diskUsed = getFilesystemUsed(propertyName+strlen(PROPERTY_DISK_USED));
		sprintf(buf, "%i", diskUsed);
		return buf;
	}
	else if (strlen(propertyName)>strlen(PROPERTY_DISK_FREE) && strncmp(propertyName,PROPERTY_DISK_FREE,strlen(PROPERTY_DISK_FREE))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int diskFree = getFilesystemFree(propertyName+strlen(PROPERTY_DISK_FREE));
		sprintf(buf, "%i", diskFree);
		return buf;
	}
	else if (strlen(propertyName)>strlen(PROPERTY_DISK_LOAD) && strncmp(propertyName,PROPERTY_DISK_LOAD,strlen(PROPERTY_DISK_LOAD))==0)
	{
		char *buf = (char *)calloc(5, sizeof(char));
		short diskLoad = getFilesystemLoad(propertyName+strlen(PROPERTY_DISK_LOAD));
		sprintf(buf, "%i", diskLoad);
		return buf;
	}
	
	else if (strlen(propertyName)>strlen(PROPERTY_NET_ADDR) && strncmp(propertyName,PROPERTY_NET_ADDR,strlen(PROPERTY_NET_ADDR))==0)
		return getNetworkIpAddress(propertyName+strlen(PROPERTY_NET_ADDR));
	else if (strlen(propertyName)>strlen(PROPERTY_NET_DATA_IN) && strncmp(propertyName,PROPERTY_NET_DATA_IN,strlen(PROPERTY_NET_DATA_IN))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int dataIn = getNetworkDataIn(propertyName+strlen(PROPERTY_NET_DATA_IN));
		sprintf(buf, "%i", dataIn);
		return buf;
	}
	else if (strlen(propertyName)>strlen(PROPERTY_NET_DATA_OUT) && strncmp(propertyName,PROPERTY_NET_DATA_OUT,strlen(PROPERTY_NET_DATA_OUT))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int dataOut = getNetworkDataOut(propertyName+strlen(PROPERTY_NET_DATA_OUT));
		sprintf(buf, "%i", dataOut);
		return buf;
	}
	else if (strlen(propertyName)>strlen(PROPERTY_NET_SPEED_IN) && strncmp(propertyName,PROPERTY_NET_SPEED_IN,strlen(PROPERTY_NET_SPEED_IN))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int speedIn = getNetworkSpeedIn(propertyName+strlen(PROPERTY_NET_SPEED_IN));
		sprintf(buf, "%i", speedIn);
		return buf;
	}
	else if (strlen(propertyName)>strlen(PROPERTY_NET_SPEED_OUT) && strncmp(propertyName,PROPERTY_NET_SPEED_OUT,strlen(PROPERTY_NET_SPEED_OUT))==0)
	{
		char *buf = (char *)calloc(16, sizeof(char));
		int speedOut = getNetworkSpeedOut(propertyName+strlen(PROPERTY_NET_SPEED_OUT));
		sprintf(buf, "%i", speedOut);
		return buf;
	}
	
	else
	{
		string msg = "Plugin_SysInfo: Invalid property '";
		msg.append(propertyName);
		msg.append("'\n");
		tuxlcdPtr->log(LOG_LEVEL_WARN, msg);
		return NULL;
	}
}


/**
 * Outputs all properties and their values to stdout.
 * Used for testing and debugging.
 */
void Plugin_SysInfo::dumpProperties()
{
	cout << "Plugin_SysInfo: Dumping properties..." << endl;
	vector<string> props = getProperties();
	for (size_t i=0; i<props.size(); i++)
	{
		cout << "  " << props[i] << "\t\t\t" << getProperty(props[i].c_str()) << "\n";
	}
}


/**
 * Returns the current time in the format %H:%M:%S.
 * \return the current time.
 */
char* Plugin_SysInfo::getTime()
{
	time_t result1;
	tm *result2;
	char *timeStr = (char *)calloc(9, sizeof(char));
	
	result1 = time(NULL);
	result2 = localtime(&result1); 
	strftime(timeStr, 9, "%H:%M:%S", result2);
	
	return timeStr;
}


/**
 * Returns the current date in the format %a %d %b %Y.
 * \return the current date.
 */
char* Plugin_SysInfo::getDate()
{
	time_t result1;
	tm *result2;
	char *dateStr = (char *)calloc(16, sizeof(char));
	
	result1 = time(NULL);
	result2 = localtime(&result1); 
	strftime(dateStr, 16, "%a %d %b %Y", result2);
	
	return dateStr;
}


/**
 * Returns the CPU load in %.
 * \return the CPU load in %.
 */
short Plugin_SysInfo::getCpuLoad()
{
	glibtop_cpu prevCpu, cpu;
	unsigned int user_diff, nice_diff, system_diff, idle_diff, iowait_diff, irq_diff, softirq_diff;
	short load;
	
	glibtop_get_cpu(&prevCpu);
	sleep(1);
	glibtop_get_cpu(&cpu);
	
	user_diff = cpu.user - prevCpu.user;
	nice_diff = cpu.nice - prevCpu.nice;
	system_diff = cpu.sys - prevCpu.sys;
	idle_diff = cpu.idle - prevCpu.idle;
	iowait_diff = cpu.iowait - prevCpu.iowait;
	irq_diff = cpu.irq - prevCpu.irq;
	softirq_diff = cpu.softirq - prevCpu.softirq;
	
	load = (short)round((100.0f / (user_diff + nice_diff + system_diff + idle_diff + iowait_diff + irq_diff + softirq_diff)) * (user_diff + nice_diff + system_diff + irq_diff + softirq_diff));

	return load;
}


/**
 * Returns the total amount of memory in megabytes.
 * \return the total amount of memory in megabytes.
 */
int Plugin_SysInfo::getMemoryTotal()
{
	glibtop_mem memory;
	glibtop_get_mem(&memory);
	return memory.total/(1024*1024);
}

/**
 * Returns the amount of memory used in megabytes.
 *  \return the amount of memory used in megabytes.
 */
int Plugin_SysInfo::getMemoryUsed()
{
	glibtop_mem memory;
	glibtop_get_mem(&memory);
	return memory.used/(1024*1024);
}


/**
 * Returns the amount of free memory in megabytes.
 * \return the amount of free memory in megabytes.
 */
int Plugin_SysInfo::getMemoryFree()
{
	glibtop_mem memory;
	glibtop_get_mem(&memory);
	return memory.free/(1024*1024);
}


/**
 * Returns the memory usage in %.
 * \return the memory usage in %.
 */
short Plugin_SysInfo::getMemoryLoad()
{
	glibtop_mem memory;
	glibtop_get_mem(&memory);
	return (short)round((memory.used / (float)memory.total) * 100.0f);
}


/**
 * Returns a list of mountpoints.
 * \return a list of mountpoints.
 */
vector<string> Plugin_SysInfo::getFilesystems()
{
	glibtop_mountlist mountList;
	glibtop_mountentry *mountEntries;
	size_t i;
	
	mountEntries = glibtop_get_mountlist(&mountList, FALSE);
	vector<string> fileSystems(mountList.number);
	for (i=0; i<mountList.number; i++)
		fileSystems[i] = mountEntries[i].mountdir;
	
	g_free(mountEntries);
	return fileSystems;
}


/**
 * Returns the size of the specified filesystem in megabytes.
 * \param[in] fileSystem the filesystem mount point.
 * \return the size of the filesystem in megabytes.
 */
int Plugin_SysInfo::getFilesystemTotal(const char *fileSystem)
{
	glibtop_fsusage fsusage;
	glibtop_get_fsusage(&fsusage, fileSystem);
	return (fsusage.blocks * fsusage.block_size >> 20);
	
}


/**
 * Returns the used space in megabytes on the specified filesystem.
 * \param[in] fileSystem the filesystem mount point.
 * \return the used space in megabytes.
 */
int Plugin_SysInfo::getFilesystemUsed(const char *fileSystem)
{
	glibtop_fsusage fsusage;
	glibtop_get_fsusage(&fsusage, fileSystem);
	return ((fsusage.blocks - fsusage.bavail) * fsusage.block_size >> 20);
}


/**
 * Returns the free space in megabytes on the specified filesystem.
 * \param[in] fileSystem the filesystem mount point.
 * \return the free space in megabytes.
 */
int Plugin_SysInfo::getFilesystemFree(const char *fileSystem)
{
	glibtop_fsusage fsusage;
	glibtop_get_fsusage(&fsusage, fileSystem);
	return (fsusage.bavail * fsusage.block_size >> 20);
}


/**
 * Returns the percentage of used space on the filesystem.
 * \param[in] fileSystem the filesystem mount point.
 * \return the percentage of used space on the filesystem.
 */
short Plugin_SysInfo::getFilesystemLoad(const char *fileSystem)
{
	glibtop_fsusage fsusage;
	glibtop_get_fsusage(&fsusage, fileSystem);
	return (short)round((fsusage.blocks - fsusage.bavail) * 100.0 / fsusage.blocks);
}


/**
 * Returns a list of network interfaces.
 * \return a list of network interfaces.
 */
vector<string> Plugin_SysInfo::getNetworkInterfaces()
{
	glibtop_netlist netList;
	size_t i;
	
	char **devices = glibtop_get_netlist(&netList);
	vector<string> interfaces(netList.number);
	for(i=0; i<netList.number; i++)
		interfaces[i] = devices[i];
	g_strfreev(devices);
	
	return interfaces;
}


/**
 * Returns the IP address of the specified interface.
 * \param[in] interface.
 * \return  address of the specified interface.
 */
char* Plugin_SysInfo::getNetworkIpAddress(const char *interface)
{
	glibtop_netload netLoad;
	struct in_addr addr;
	
	glibtop_get_netload(&netLoad, interface);
	addr.s_addr = netLoad.address;
	return inet_ntoa(addr);
}


/**
 * Returns the total number of megabytes received.
 * \param[in] interface.
 * \return the total number of megabytes received.
 */
int Plugin_SysInfo::getNetworkDataIn(const char *interface)
{
	glibtop_netload netLoad;
	glibtop_get_netload(&netLoad, interface);
	return netLoad.bytes_in/(1024*1024);
}


/**
 * Returns the total number of megabytes transmitted.
 * \param[in] interface.
 * \return the total number of megabytes transmitted.
 */
int Plugin_SysInfo::getNetworkDataOut(const char *interface)
{
	glibtop_netload netLoad;
	glibtop_get_netload(&netLoad, interface);
	return netLoad.bytes_out/(1024*1024);
}


/**
 * Returns the current download speed in kb/sec.
 * \param[in] interface.
 * \return the current download speed in kb/sec.
 */
int Plugin_SysInfo::getNetworkSpeedIn(const char *interface)
{
	glibtop_netload netLoad;
	int prevBytesIn, bytesIn;
	
	glibtop_get_netload(&netLoad, interface);
	prevBytesIn = netLoad.bytes_in;
	sleep(1);
	glibtop_get_netload(&netLoad, interface);
	bytesIn = netLoad.bytes_in;
	
	return (bytesIn-prevBytesIn)/1024;
}


/**
 * Returns the current upload speed in kb/sec.
 * \param[in] interface.
 * \return the current upload speed in kb/sec.
 */
int Plugin_SysInfo::getNetworkSpeedOut(const char *interface)
{
	glibtop_netload netLoad;
	int prevBytesOut, bytesOut;
	
	glibtop_get_netload(&netLoad, interface);
	prevBytesOut = netLoad.bytes_out;
	sleep(1);
	glibtop_get_netload(&netLoad, interface);
	bytesOut = netLoad.bytes_out;
	
	return (bytesOut-prevBytesOut)/1024;
}
