#include <string>		// String function definitions
#include <vector>		//
#include <sys/utsname.h>	//
#include "../TuxLCD.h"
#include "../Plugin.h"
using namespace std;

#define PROPERTY_TIME		"TIME"
#define PROPERTY_DATE		"DATE"

#define PROPERTY_SYSNAME	"SYSNAME"
#define PROPERTY_NODENAME	"NODENAME"
#define PROPERTY_RELEASE	"RELEASE"
#define PROPERTY_VERSION	"VERSION"
#define PROPERTY_MACHINE	"MACHINE"

#define PROPERTY_CPU_LOAD	"CPU_LOAD"
#define PROPERTY_MEMORY_TOTAL	"MEMORY_TOTAL"
#define PROPERTY_MEMORY_USED	"MEMORY_USED"
#define PROPERTY_MEMORY_FREE	"MEMORY_FREE"
#define PROPERTY_MEMORY_LOAD	"MEMORY_LOAD"

#define PROPERTY_DISK_TOTAL	"DISK_TOTAL_"
#define PROPERTY_DISK_USED	"DISK_USED_"
#define PROPERTY_DISK_FREE	"DISK_FREE_"
#define PROPERTY_DISK_LOAD	"DISK_LOAD"

#define PROPERTY_NET_ADDR	"NET_ADDR_"
#define PROPERTY_NET_DATA_IN	"NET_DATA_IN_"
#define PROPERTY_NET_DATA_OUT	"NET_DATA_OUT_"
#define PROPERTY_NET_SPEED_IN	"NET_SPEED_IN_"
#define PROPERTY_NET_SPEED_OUT	"NET_SPEED_OUT_"


/**
 * System information plugin for TuxLCD.
 *
 * \author David Bull
 * \version 0.1
 * \date Monday 7th August 2006
 */
class Plugin_SysInfo : public Plugin
{
	private:
		TuxLCD *tuxlcdPtr;
		struct utsname name;
		
		
		/**
		 * Returns the current time in the format %H:%M:%S.
		 * \return the current time.
		 */
		char* getTime();
		
		
		/**
		 * Returns the current date in the format %a %d %b %Y.
		 * \return the current date.
		 */
		char* getDate();
		
		
		/**
		 * Returns the CPU usage in %.
		 * \return the CPU usage in %.
		 */
		short getCpuLoad();
		
		
		/**
		 * Returns the total amount of memory in megabytes.
		 * \return the total amount of memory in megabytes.
		 */
		int getMemoryTotal();
		
		/**
		 * Returns the amount of memory used in megabytes.
		 * \return the amount of memory used in megabytes.
		 */
		int getMemoryUsed();
		
		
		/**
		 * Returns the amount of free memory in megabytes.
		 * \return the amount of free memory in megabytes.
		 */
		int getMemoryFree();
		
		
		/**
		 * Returns the memory usage in %.
		 * \return the memory usage in %.
		 */
		short getMemoryLoad();
		
		
		/**
		 * Returns a list of mountpoints.
		 * \return a list of mountpoints.
		 */
		vector<string> getFilesystems();
		
		
		/**
		 * Returns the size of the specified filesystem in megabytes.
		 * \param[in] fileSystem the filesystem mount point.
		 * \return the size of the filesystem in megabytes.
		 */
		int getFilesystemTotal(const char *fileSystem);
		
		
		/**
		 * Returns the used space in megabytes on the specified filesystem.
		 * \param[in] fileSystem the filesystem mount point.
		 * \return the used space in megabytes.
		 */
		int getFilesystemUsed(const char *fileSystem);
		
		
		/**
		 * Returns the free space in megabytes on the specified filesystem.
		 * \param[in] fileSystem the filesystem mount point.
		 * \return the free space in megabytes.
		 */
		int getFilesystemFree(const char *fileSystem);
		
		
		/**
		 * Returns the percentage of used space on the filesystem.
		 * \param[in] fileSystem the filesystem mount point.
		 * \return the percentage of used space on the filesystem.
		 */
		short getFilesystemLoad(const char *fileSystem);
		
		
		/**
		 * Returns a list of network interfaces.
		 * \return a list of network interfaces.
		 */
		vector<string> getNetworkInterfaces();
		
		
		/**
		 * Returns the IP address of the specified interface.
		 * \param[in] interface.
		 * \return  address of the specified interface.
		 */
		char* getNetworkIpAddress(const char *interface);
		
		
		/**
		 * Returns the total number of megabytes received.
		 * \param[in] interface.
		 * \return the total number of megabytes received.
		 */
		int getNetworkDataIn(const char *interface);
		
		
		/**
		 * Returns the total number of megabytes transmitted.
		 * \param[in] interface.
		 * \return the total number of megabytes transmitted.
		 */
		int getNetworkDataOut(const char *interface);
		
		
		/**
		 * Returns the current download speed in kb/sec.
		 * \param[in] interface.
		 * \return the current download speed in kb/sec.
		 */
		int getNetworkSpeedIn(const char *interface);
		
		
		/**
		 * Returns the current upload speed in kb/sec.
		 * \param[in] interface.
		 * \return the current upload speed in kb/sec.
		 */
		int getNetworkSpeedOut(const char *interface);
	
	
	public:
		/**
		 * Constructor.
		 * \param[in] tuxlcd pointer back to tuxlcd for callbacks such as when the keypad on the LCD is pressed.
		 */
		Plugin_SysInfo(TuxLCD *tuxlcd);
		
		
		/**
		 * Destrustor.
		 */
		~Plugin_SysInfo();
		
		
		/**
		 * Returns the name of this plugin.
		 * \return the plugin name.
		 */
		const char* getPluginName()
		{
			return "SysInfo";
		}
		
		
		/**
		 * Returns a list of properties available from this plugin.
		 * Property values can then be retrieved by calling getProperty(propertyName).
		 * \see getProperty(propertyName).
		 * \return a list of properties available from this plugin.
		 */
		vector<string> getProperties();
		
		
		/**
		 * Returns the value of the specified property, or NULL if the propertyName is not valid.
		 * A list of valid property names is available from getProperties().
		 * \see getProperties()
		 * \param[in] propertyName the name of the property.
		 * \return the value of the specified property, or NULL if the propertyName is not valid.
		 */
		char* getProperty(const char *propertyName);
		
		
		/**
		 * Outputs all properties and their values to stdout.
		 * Used for testing and debugging.
		 */
		void dumpProperties();
};


// the class factories

extern "C" Plugin* create(void *tuxlcd)
{
	return new Plugin_SysInfo((TuxLCD*)tuxlcd);
}

extern "C" void destroy(Plugin *p)
{
	delete p;
}
